
import sys
import inspect
import builtins
import importlib
from fractions import Fraction
from decimal import Decimal
from typing import Any, List, Dict, Callable


ZERO_PARAMS: List[str] = ['args', 'kwargs']


def simple_serialization(class_: type) -> type:
    '''A decorator to provide a simple to_dict() serialization method.

    The resulting method will serialize all object attributes corresponding
    to the class's constructor parameter names. Therefore, this decorator
    is only useful when the class stores all its original parameters
    unchanged (or in any other form acceptable to its constructor).

    :param class_: The class to add the method to.
    '''
    if hasattr(class_, 'serialize_params'):
        param_names = class_.serialize_params
    else:
        param_names = list(inspect.signature(
            class_.__init__
        ).parameters.keys())
        if 'self' in param_names:
            param_names.remove('self')
        if param_names == ZERO_PARAMS and class_.__init__ == object.__init__:
            param_names = []

    def to_dict(self) -> Dict[str, Any]:
        out_dict = {'class': scoped_class_name(self)}
        for attr in param_names:
            out_dict[attr] = serialize_value(getattr(self, attr))
        return out_dict

    class_.to_dict = to_dict
    return class_


def serialize_value(value: Any) -> Any:
    if hasattr(value, 'to_dict'):
        return value.to_dict()
    elif isinstance(value, tuple(ATOMIC_TYPES)):
        return value
    elif type(value) in CONVERTIBLE_TYPES:
        return CONVERTIBLE_TYPES[type(value)](value)
    elif hasattr(value, '__iter__'):
        if hasattr(value, 'items') and hasattr(value, 'keys'):
            if all(isinstance(key, str) for key in value.keys()):
                return {
                    key: serialize_value(val)
                    for key, val in value.items()
                }
            else:
                return {
                    'type': 'dict',
                    'keys': [serialize_value(key) for key in value.keys()],
                    'values': [serialize_value(val) for val in value.values()]
                }
        else:
            return [serialize_value(val) for val in value]
    elif hasattr(value, '__call__'):
        return {'callable': '.'.join((value.__module__, value.__name__))}
    else:
        raise ValueError(f'cannot serialize {value!r} to dict format')


def deserialize_value(value: Any) -> Any:
    if isinstance(value, dict):
        if 'type' in value and is_scoped_identifier(value['type']):
            return deserialize_typed(value)
        elif 'class' in value and is_scoped_identifier(value['class']):
            return deserialize_class(value)
        elif 'callable' in value and is_scoped_identifier(value['callable']):
            return get_object(value['callable'])
        else:
            return {key: deserialize_value(val) for key, val in value.items()}
    elif isinstance(value, tuple(ATOMIC_TYPES)):
        return value
    elif isinstance(value, list):
        return [deserialize_value(val) for val in value]
    else:
        raise ValueError(f'cannot deserialize {value!r}, type unknown')


def deserialize_typed(typedef: Dict[str, Any]) -> Any:
    typeobj = get_object(typedef['type'])
    if typeobj is dict:
        return dict(zip(
            [deserialize_value(key) for key in typedef['keys']],
            [deserialize_value(val) for val in typedef['values']]
        ))
    elif 'value' in typedef:
        return typeobj(deserialize_value(typedef['value']))
    elif 'arguments' in typedef:
        return typeobj(*[
            deserialize_value(val) for val in typedef['arguments']
        ])
    elif 'parameters' in typedef:
        return typeobj(**{
            param: deserialize_value(val)
            for param, val in typedef['parameters'].items()
        })
    else:
        raise ValueError(f'invalid typed value contents: {typedef!r}')


def deserialize_class(clsdef: Dict[str, Any]) -> Any:
    cls = get_object(clsdef['class'])
    params = clsdef.copy()
    del params['class']
    if hasattr(cls, 'from_dict'):
        return cls.from_dict(params)
    else:
        for key, inner_val in params.items():
            params[key] = deserialize_value(inner_val)
        return cls(**params)


def get_object(identifier: str) -> Any:
    if '.' not in identifier:
        global_vars = globals()
        if identifier in global_vars:
            return global_vars[identifier]
        else:
            return getattr(builtins, identifier)
    else:
        module, name = identifier.rsplit('.', 1)
        if module not in sys.modules:
            importlib.import_module(module)
        return getattr(sys.modules[module], name)


def from_dict(value: Dict[str, Any]) -> Any:
    """Parse an election evaluator object from a JSON-like dictionary.

    :param value: A dictionary created by :func:`to_dict`.
    """
    if not isinstance(value, dict):
        raise ValueError('invalid votelib object def: dict expected,'
                         f'got {value!r}')
    elif 'class' not in value:
        raise ValueError('invalid votelib object def: must have a class key')
    elif not is_scoped_identifier(value['class']):
        inval_cls = value['class']
        raise ValueError(f"invalid votelib class def: {inval_cls}")
    else:
        return deserialize_value(value)


def to_dict(obj: Any) -> Dict[str, Any]:
    """Serialize an election evaluator object to a JSON-ready dictionary.

    :param obj: An election evaluator object or similar. It should provide
        a `to_dict()` method (all the standard evaluators, converters and the
        like from Votelib should have it, courtesy of the simple_serialization
        decorator).
    """
    return serialize_value(obj)


def is_scoped_identifier(value: Any):
    return (
        isinstance(value, str)
        and not value.startswith('.')
        and all(chunk.isidentifier() for chunk in value.split('.'))
    )


def scoped_class_name(value: Any):
    cls = value.__class__
    return '.'.join((cls.__module__, cls.__name__))


def fraction_to_json(f: Fraction) -> Dict[str, Any]:
    return {'type': 'Fraction', 'arguments': f.as_integer_ratio()}


def decimal_to_json(d: Decimal) -> Dict[str, Any]:
    return {'type': 'Decimal', 'value': str(d)}


def sequence_to_json_factory(typeobj):
    typename = typeobj.__name__

    def sequence_to_json(seq) -> Dict[str, Any]:
        return {'type': typename, 'value': [serialize_value(v) for v in seq]}

    return sequence_to_json


ATOMIC_TYPES: List[type] = [
    str, int, float, bool, type(None),
]

CONVERTIBLE_TYPES: Dict[type, Callable] = {
    Fraction: fraction_to_json,
    Decimal: decimal_to_json,
}

SEQUENCE_TYPES: List[type] = [frozenset, tuple]

for seqtype in SEQUENCE_TYPES:
    CONVERTIBLE_TYPES[seqtype] = sequence_to_json_factory(seqtype)




from decimal import Decimal
from numbers import Number
from typing import List, Dict, Tuple, Set, Iterable, Optional

import votelib.candidate
import votelib.util
import votelib.io.core
from votelib.candidate import Candidate


BLTSpecContents = Tuple[
    Dict[Tuple[Candidate, ...], Number],    # ranked votes w/o shared ranks
    int,
    List[Candidate],
    Optional[str],
]


class NotSupportedInBLT(votelib.io.core.NotSupportedInFormat):
    FORMAT = 'BLT file'


class BLTParseError(votelib.io.core.ParseError):
    pass


def dump_lines(votes: Dict[Tuple[Candidate, ...], Number],
               n_seats: int,
               candidates: Optional[List[Candidate]] = None,
               election_name: Optional[str] = None,
               ) -> Iterable[str]:
    if candidates is None:
        candidates = votelib.util.all_ranked_candidates(votes)
    yield _dump_numline([len(candidates), n_seats])
    for i in _get_withdrawn_inds(candidates):
        yield _dump_numline([-(i+i)])
    for rvote, n_votes in votes.items():
        yield _dump_numline(_dump_vote(rvote, candidates, n_votes))
    yield _dump_numline([0])
    for cand in candidates:
        if not isinstance(cand, str):
            if hasattr(cand, 'name'):
                cand = cand.name
            else:
                cand = str(cand)
        yield _dump_strline(cand)
    if election_name is not None:
        yield _dump_strline(election_name)


dump, dumps = votelib.io.core.dumpers(dump_lines)


def _get_withdrawn_inds(candidates: List[Candidate]) -> List[int]:
    return [
        i for i, cand in enumerate(candidates)
        if hasattr(cand, 'withdrawn') and cand.withdrawn
    ]


def _dump_vote(vote: Tuple[Candidate, ...],
               candidates: List[Candidate],
               n_votes: Number,
               ) -> List[Number]:
    try:
        cand_indices = [candidates.index(cand) + 1 for cand in vote]
    except ValueError:
        raise NotSupportedInBLT(f'equal rankings: {vote}')
    else:
        return [n_votes] + cand_indices + [0]


def _dump_numline(nums: List[Number]) -> str:
    return ' '.join(str(num) for num in nums)


def _dump_strline(string: str) -> str:
    return f'"{string}"'


def load_lines(blt_lines: Iterable[str],
               oneplus_weights: bool = False,
               ) -> BLTSpecContents:
    try:
        n_cands, n_seats = _parse_header(next(blt_lines))
    except StopIteration as e:
        raise BLTParseError('empty BLT file') from e
    ballots, withdrawn = _parse_body(
        blt_lines,
        oneplus_weights=oneplus_weights
    )
    candidates, election_name = _parse_strings(blt_lines, n_cands)
    if candidates is None:
        candidates = _numeric_candidates(n_cands)
    candidates = _form_candidate_objects(candidates, withdrawn)
    true_ballots = _deindex_ballots(ballots, candidates)
    return (
        true_ballots,
        n_seats,
        candidates,
        election_name,
    )


load, loads = votelib.io.core.loaders(load_lines)


def _numeric_candidates(n_cands: int) -> List[str]:
    return [str(i+1) for i in range(n_cands)]


def _form_candidate_objects(cands: List[str],
                            withdrawn: Set[int],
                            ) -> List[votelib.candidate.Person]:
    return [
        votelib.candidate.Person(
            cand,
            number=i+1,
            withdrawn=(i+1 in withdrawn)
        )
        for i, cand in enumerate(cands)
    ]


def _deindex_ballots(ballots: Dict[Tuple[int, ...], Number],
                     cands: List[Candidate]
                     ) -> Dict[Tuple[Candidate, ...], Number]:
    return {
        tuple(cands[i-1] for i in ballot): n_votes
        for ballot, n_votes in ballots.items()
    }


def _parse_header(blt_line: str) -> Tuple[int, int]:
    blt_result = _parse_numline(blt_line, allow_first_decimal=False)
    is_ok = (
        isinstance(blt_result, list)
        and len(blt_result) == 2
        and all(isinstance(num, int) for num in blt_result)
    )
    if is_ok:
        return tuple(blt_result)
    else:
        raise BLTParseError(f'need two integers (candidate and seat count)'
                            f' in BLT file header line, got {blt_result!r}')


def _parse_body(blt_lines: Iterable[str],
                oneplus_weights: bool = False,
                ) -> Tuple[Dict[Tuple[int, ...], Number], Set[int]]:
    ballots = {}
    withdrawn = set()
    ballots_encountered = False
    for line in blt_lines:
        result = _parse_numline(line, allow_first_decimal=True)
        if not result:
            continue    # ignore empty lines
        elif result == [0]:
            # End-of-ballots line, return.
            return ballots, withdrawn
        elif result[0] < 0:
            if ballots_encountered:
                raise BLTParseError('withdrawn candidate line after ballot'
                                    f' line: {line!r}')
            # Withdrawn candidates. Allow more than one per line.
            withdrawn.update(-n for n in result)
        else:
            weight, ballot = _parse_ballot(result)
            if oneplus_weights and weight < 1:
                raise ValueError(f'ballot weight <1: {line!r}')
            if ballot not in ballots:
                ballots[ballot] = 0
            ballots[ballot] += weight
            ballots_encountered = True
    raise BLTParseError('incomplete BLT file:'
                        ' EOF before ballot list terminator')


def _parse_strings(blt_lines: Iterable[str],
                   n_cands: int,
                   ) -> Tuple[Optional[List[str]], Optional[str]]:
    parsed_lines = []
    empty_encountered = False
    for blt_line in blt_lines:
        blt_line = _clean_line(blt_line)
        if blt_line.startswith('"') and blt_line.endswith('"'):
            if empty_encountered:
                raise BLTParseError(f'nonempty line after empty: {blt_line!r}')
            parsed_lines.append(blt_line[1:-1])
        elif not blt_line:
            empty_encountered = True
        else:
            raise BLTParseError(f'invalid BLT string line: {blt_line!r}')
    if not parsed_lines:
        return None, None
    if len(parsed_lines) == 1:
        if n_cands == 1:
            return parsed_lines[0], None
        else:
            return None, parsed_lines[0]
    elif len(parsed_lines) < n_cands:
        raise BLTParseError(f'not enough candidate names: {len(parsed_lines)}'
                            f' given, {n_cands} set in header')
    elif len(parsed_lines) == n_cands:
        return parsed_lines, None
    elif len(parsed_lines) == n_cands + 1:
        return parsed_lines[:-1], parsed_lines[-1]
    else:
        raise BLTParseError(f'too many strings: {len(parsed_lines)} found'
                            f'but expecting {n_cands} candidate names + title')


def _clean_line(blt_line: str) -> str:
    blt_line = blt_line.strip()
    # Ignore everything after the first hash sign after the last double quote.
    hash_search_start = blt_line.rfind('"') if '"' in blt_line else 0
    leftmost_hash = blt_line[hash_search_start:].find('#')
    if leftmost_hash == -1:
        return blt_line
    else:
        return blt_line[:(hash_search_start + leftmost_hash)].rstrip()


def _parse_ballot(nums: List[Number]) -> Tuple[Number, Tuple[int, ...]]:
    # Assumes a line with at least one leading non-zero element, all elements
    # apart from the first one are assumed to be integers.
    # Check the trailing zero and strip it.
    if nums[-1] != 0:
        raise BLTParseError('ballot line must be zero-terminated,'
                            f'got {nums!r}')
    nums = nums[:-1]
    # The first element is weight, the rest are candidate indices.
    return nums[0], tuple(nums[1:])


def _parse_numline(blt_line: str,
                   allow_first_decimal: bool = False,
                   ) -> List[Number]:
    blt_line = _clean_line(blt_line)
    # Return empty lines as None.
    if not blt_line:
        return []
    # Split the line by spaces to obtain numbers.
    nums = []
    for i, numstr in enumerate(blt_line.split()):
        if numstr.isdigit():
            nums.append(int(numstr))
        elif i == 0 and allow_first_decimal:
            nums.append(Decimal(numstr))
        else:
            raise BLTParseError(f'invalid BLT numberline item {i}: {numstr!r}'
                                f'(first decimal item'
                                f'allowed: {allow_first_decimal})')
    return nums



'''Distribution evaluators that are usually called proportional.

This contains the most common distribution evaluators used in party-list
elections that at least aim to be proportional (although some parameter setups
might make the result very nonproportional, such as the Imperiali divisor)
and also some similar, simpler distribution evaluators that aim for strict
proportionality while relaxing some of the constraints of a distribution
evaluator.
'''

import bisect
import collections
import decimal
from fractions import Fraction
from typing import (
    Any, List, Tuple, Dict, Union, Callable, Optional, Set, Collection
)
from numbers import Number

import votelib.convert
import votelib.util
import votelib.component.divisor
import votelib.component.quota
import votelib.evaluate.core
from votelib.candidate import Candidate, Constituency
from votelib.persist import simple_serialization

INF = float('inf')


@simple_serialization
class PureProportionality:
    '''Distribute seats among candidates strictly proportionally (no rounding).

    This evaluator is mostly auxiliary since it gives fractional seat counts.
    '''
    def evaluate(self,
                 votes: Dict[Candidate, int],
                 n_seats: int,
                 prev_gains: Dict[Candidate, int] = {},
                 max_seats: Dict[Candidate, int] = {},
                 ) -> Dict[Candidate, Fraction]:
        '''Distribute seats exactly proportionally, giving fractional seats.

        :param votes: Simple votes to be evaluated.
        :param n_seats: Number of seats to be filled.
        :param prev_gains: Seats gained by the candidate/party in previous
            election rounds to be subtracted from the proportional result
            awarded here.
        :param max_seats: Maximum number of seats that the given
            candidate/party can obtain in total (including previous gains).
        '''
        fixed = []
        result = {}
        prev_len_fixed = -1
        while len(fixed) != prev_len_fixed:
            prev_len_fixed = len(fixed)
            result = {
                cand: cseats for cand, cseats in result.items()
                if cand in fixed
            }
            budget = n_seats - sum(result.values())
            current_votes = {
                cand: n_votes for cand, n_votes in votes.items()
                if cand not in fixed
            }
            seats_per_vote = Fraction(budget, sum(current_votes.values()))
            for cand, n_votes in current_votes.items():
                cand_give_seats = n_votes * seats_per_vote
                if int(cand_give_seats) == cand_give_seats:
                    cand_give_seats = int(cand_give_seats)
                cand_has_seats = prev_gains.get(cand, 0)
                cand_max_seats = max_seats.get(cand, INF)
                if cand_give_seats > cand_has_seats:
                    if cand_give_seats > cand_max_seats:
                        # proportional result over maximum, fix maximum
                        fixed.append(cand)
                        cand_give_seats = cand_max_seats
                    result[cand] = cand_give_seats
                else:
                    # proportional result below minimum, fix minimum
                    fixed.append(cand)
                    result[cand] = cand_has_seats
        return {
            cand: cseats - prev_gains.get(cand, 0)
            for cand, cseats in result.items()
        }


@simple_serialization
class VotesPerSeat:
    """Award seats for each N votes cast for each candidate.

    This is an old and simple system that was used e.g. in pre-war Germany
    [#wrstag]_. It divides the number of votes by a pre-specified constant and
    rounds to give the appropriate number of seats. It is also used as an
    auxiliary evaluator in some other systems with fixed quota.

    :param votes_per_seat: Number of votes required for the candidate to
        obtain a single seat.
    :param rounding: A rounding mode from the *decimal* Python library. The
        default option is to round down, which is the most frequent case,
        but this evaluator allows to specify a different method as well.
    :param accept_equal: If False, whenever the number of votes is exactly
        divisible by *votes_per_seat*, award one less seat.

    .. [#wrstag] "Reichstag (Weimarer Republik): Wahlsystem", Wikipedia.
        https://de.wikipedia.org/wiki/Reichstag_(Weimarer_Republik)#Wahlsystem
    """
    def __init__(self,
                 votes_per_seat: int,
                 rounding: str = decimal.ROUND_DOWN,
                 accept_equal: bool = True,
                 ):
        self.votes_per_seat = votes_per_seat
        self.rounding = rounding
        self.accept_equal = accept_equal

    def evaluate(self,
                 votes: Dict[Candidate, int],
                 prev_gains: Dict[Candidate, int] = {},
                 max_seats: Dict[Candidate, int] = {},
                 ) -> Dict[Candidate, int]:
        calc_entitlement = self._get_entitlement_computer(sum(votes.values()))
        out = {}
        for cand, n_votes in votes.items():
            entitlement = calc_entitlement(n_votes)
            if not self.accept_equal and entitlement > 0:
                if entitlement * self.votes_per_seat == n_votes:
                    entitlement -= 1
            entitlement = min(entitlement, max_seats.get(cand, INF))
            entitlement -= prev_gains.get(cand, 0)
            if entitlement:
                out[cand] = entitlement
        return out

    def _get_entitlement_computer(self,
                                  total_votes: Number,
                                  ) -> Callable[[int], int]:
        is_fractional = (
            (isinstance(total_votes, int) or isinstance(total_votes, Fraction))
            and (
                isinstance(self.votes_per_seat, int)
                or isinstance(self.votes_per_seat, Fraction)
            )
        )
        if is_fractional and self.rounding == decimal.ROUND_DOWN:
            return self._fractional_entitlement
        else:
            return self._decimal_entitlement

    def _fractional_entitlement(self, n_votes: int) -> int:
        return int(Fraction(n_votes, self.votes_per_seat))

    def _decimal_entitlement(self, n_votes: int) -> int:
        with decimal.localcontext() as context:
            context.rounding = self.rounding
            vps = decimal.Decimal(self.votes_per_seat)
            return int(round(decimal.Decimal(n_votes) / vps, 0))


@simple_serialization
class QuotaDistributor:
    '''Distribute seats proportionally, according to multiples of quota filled.

    Each contestant is awarded the number of seats according to the number
    of times their votes fill the provided quota. (This essentially means the
    numbers of votes are divided by the quota and rounded down to get the
    numbers of seats.)

    This is usually not a self-standing evaluator because it (except for very
    rare cases) does not award the full number of seats; it is, however, used
    in initial stages of some proportional systems, such as the Czech
    Chamber of Deputies election.

    :param quota_function: A callable producing the quota threshold from the
        total number of votes and number of seats. The common quota functions
        can be referenced by string name from the
        :mod:`votelib.component.quota` module.
    :param accept_equal: Whether to consider the candidate elected when
        their votes exactly equal the quota.
    :param on_overaward: Some ill-conditioned but still used quotas, such as
        Imperiali, may distribute more seats than the allocated total since
        they are low. In that case, do the following:

        -   ``'ignore'`` - return the result regardless,
        -   ``'error'`` - raise a VotingSystemError,
        -   ``'subtract'`` - subtract seats from the candidates with the
            smallest remainder after the quota division (those that exceed the
            quota by a lowest margin) until the total seat count is met.
    '''
    def __init__(self,
                 quota_function: Union[
                     str, Callable[[int, int], Number]
                 ] = 'droop',
                 accept_equal: bool = True,
                 on_overaward: str = 'error',
                 ):
        self.quota_function = votelib.component.quota.construct(quota_function)
        self.accept_equal = accept_equal
        if on_overaward not in ('ignore', 'error', 'subtract'):
            raise ValueError(f'invalid on_overaward setting: {on_overaward}')
        self.on_overaward = on_overaward

    def evaluate(self,
                 votes: Dict[Candidate, int],
                 n_seats: int,
                 prev_gains: Dict[Candidate, int] = {},
                 max_seats: Dict[Candidate, int] = {},
                 ) -> Dict[Candidate, int]:
        '''Distribute seats proportionally by multiples of quota filled.

        :param votes: Simple votes to be evaluated.
        :param n_seats: Number of seats to be filled.
        :param prev_gains: Seats gained by the candidate/party in previous
            election rounds to be subtracted from the proportional result
            awarded here.
        :param max_seats: Maximum number of seats that the given
            candidate/party can obtain in total (including previous gains).
        '''
        quota_val = self.quota_function(
            sum(votes.values()), n_seats
        )
        selected = {}
        n_overshot = 0
        overshot_candidates = []
        for candidate, n_votes in votes.items():
            n_prev = prev_gains.get(candidate, 0)
            fulfills_quota = (
                n_votes > quota_val
                or self.accept_equal and n_votes == quota_val
            )
            if fulfills_quota:
                n_add_seats = int(Fraction(n_votes, quota_val)) - n_prev
                if n_add_seats > 0:
                    cand_max_seats = max_seats.get(candidate, n_seats)
                    if n_add_seats + n_prev > cand_max_seats:
                        overshoot = n_add_seats + n_prev
                        n_add_seats -= overshoot
                        n_overshot += overshoot
                        overshot_candidates.append(candidate)
                    selected[candidate] = n_add_seats
        if n_overshot:
            remaining_votes = {
                cand: n_votes for cand, n_votes in votes.items()
                if cand not in overshot_candidates
            }
            total_gained = {
                cand: selected.get(cand, 0) + prev_gains.get(cand, 0)
                for cand in votes
            }
            votelib.util.add_dict_to_dict(selected, self.evaluate(
                remaining_votes,
                n_overshot,
                prev_gains=total_gained,
                max_seats=max_seats
            ))
        total_awarded = sum(selected.values()) + sum(prev_gains.values())
        if total_awarded > n_seats:
            if self.on_overaward == 'ignore':
                return selected
            elif self.on_overaward == 'error':
                raise votelib.evaluate.core.VotingSystemError(
                    f'quota {self.quota_function.__name__} awarded total'
                    f' {total_awarded} seats, {n_seats} expected'
                )
            elif self.on_overaward == 'subtract':
                return self._subtract_overaward(
                    votes, selected, n_seats, prev_gains=prev_gains
                )
            else:
                raise ValueError
        return selected

    def _subtract_overaward(self,
                            votes: Dict[Candidate, int],
                            selected: Dict[Candidate, int],
                            n_seats: int,
                            prev_gains: Dict[Candidate, int] = {},
                            ) -> Dict[Candidate, int]:
        overaward = sum(selected.values()) + sum(prev_gains.values()) - n_seats
        quota_val = self.quota_function(
            sum(votes.values()), n_seats
        )
        while overaward > 0:
            remainders = {
                cand: -(
                    votes.get(cand, 0)
                    - quota_val * (cand_n_seats + prev_gains.get(cand, 0))
                )
                for cand, cand_n_seats in selected.items()
            }
            subtract_cand = votelib.evaluate.core.get_n_best(remainders, 1)[0]
            if isinstance(subtract_cand, votelib.evaluate.core.Tie):
                if subtract_cand in selected:
                    if selected[subtract_cand] == 1:
                        del selected[subtract_cand]
                    else:
                        selected[subtract_cand] -= 1
                else:
                    for tied_cand in subtract_cand:
                        if selected[tied_cand] == 1:
                            del selected[tied_cand]
                        else:
                            selected[tied_cand] -= 1
                    selected[subtract_cand] = (
                        selected.get(subtract_cand, 0) + len(subtract_cand) - 1
                    )
            elif selected[subtract_cand] == 1:
                del selected[subtract_cand]
            else:
                selected[subtract_cand] -= 1
            overaward -= 1
        return selected


class LargestRemainder:
    '''Distribute seats proportionally, rounding by largest remainder.

    Each contestant is awarded the number of seats according to the number
    of times their votes fill the provided quota, just like
    :class:`QuotaDistributor`. In addition to that, the
    parties that have the largest remainder of votes after the quotas are
    subtracted get an extra seat until the number of seats is exhausted.

    This includes some popular proportional party-list systems like Hare,
    Droop or Hagenbach-Bischoff. The result is usually very close to
    proportionality (closer than in highest averages systems) but might suffer
    from an Alabama paradox where adding proportionally distributed votes might
    cause one of the contestants to lose seats.

    :param quota_function: A callable producing the quota threshold from the
        total number of votes and number of seats. The common quota functions
        can be referenced by string name from the
        :mod:`votelib.component.quota` module.

    All additional keyword arguments have the same meaning as in
    :class:`QuotaDistributor`.
    '''
    def __init__(self,
                 quota_function: Union[str, Callable[[int, int], Number]],
                 **kwargs):
        self.quota_function = votelib.component.quota.construct(quota_function)
        self._quota_evaluator = QuotaDistributor(self.quota_function, **kwargs)

    def to_dict(self) -> Dict[str, Any]:
        return {
            **self._quota_evaluator.to_dict(),
            'class': votelib.persist.scoped_class_name(self)
        }

    def evaluate(self,
                 votes: Dict[Candidate, int],
                 n_seats: int,
                 prev_gains: Dict[Candidate, int] = {},
                 max_seats: Dict[Candidate, int] = {},
                 ) -> Dict[Candidate, int]:
        '''Distribute seats proportionally, rounding by largest remainder.

        :param votes: Simple votes to be evaluated.
        :param n_seats: Number of seats to be filled.
        :param prev_gains: Seats gained by the candidate/party in previous
            election rounds to be subtracted from the proportional result
            awarded here.
        :param max_seats: Maximum number of seats that the given
            candidate/party can obtain in total (including previous gains).
        '''
        # first, assign the non-remainder seats for those over the quota
        quota_elected = self._quota_evaluator.evaluate(
            votes, n_seats, prev_gains
        )
        quota_number = self.quota_function(
            sum(votes.values()), n_seats
        )
        gained_prerem = votelib.util.sum_dicts(quota_elected, prev_gains)
        n_for_remainder = n_seats - sum(gained_prerem.values())
        remainders = {
            cand: Fraction(n_votes, quota_number) - gained_prerem.get(cand, 0)
            for cand, n_votes in votes.items()
            if gained_prerem.get(cand, 0) < max_seats.get(cand, INF)
        }
        best = votelib.evaluate.core.get_n_best(remainders, n_for_remainder)
        for candidate in best:
            if candidate in quota_elected:
                quota_elected[candidate] += 1
            else:
                quota_elected[candidate] = 1
        return quota_elected


@simple_serialization
class HighestAverages:
    '''Distribute seats proportionally by ordering divided vote counts.

    Divides the vote count for each party by an increasing sequence of divisors
    (usually small integers), sorts these quotients and awards a seat for each
    of the first n_seats quotients.

    This includes some popular proportional party-list systems like D'Hondt or
    Sainte-Laguë/Webster. The result is usually quite close to proportionality
    and avoids the Alabama paradox of largest remainder systems. However, it
    usually favors either large or smaller parties, depending on the choice
    of the divisor function.

    :param divisor_function: A callable producing the divisor from the number
        of seats awarded to the contestant so far. For example, the D'Hondt
        divisor (which uses the natural numbers sequence) would always return
        the number of currently held seats raised by one. The common divisor
        functions can be referenced by string name from the
        :mod:`votelib.component.divisor` module.
    '''
    def __init__(self,
                 divisor_function: Union[
                     str, Callable[[int], Number]
                 ] = 'd_hondt',
                 ):
        self.divisor_function = votelib.component.divisor.construct(
            divisor_function
        )

    def evaluate(self,
                 votes: Dict[Candidate, int],
                 n_seats: int,
                 prev_gains: Dict[Candidate, int] = {},
                 max_seats: Dict[Candidate, int] = {},
                 ) -> Dict[Candidate, int]:
        '''Distribute seats proportionally by highest averages.

        :param votes: Simple votes to be evaluated.
        :param n_seats: Number of seats to be filled.
        :param prev_gains: Seats gained by the candidate/party in previous
            election rounds, to determine the starting divisor.
        :param max_seats: Maximum number of seats that the given
            candidate/party can obtain in total (including previous gains).
        '''
        totals = prev_gains.copy()
        quotient_dict = {}
        for cand, n_votes in votes.items():
            cand_total = totals.get(cand, 0)
            divisor = self.divisor_function(cand_total)
            if divisor > 0 and cand_total < max_seats.get(cand, n_seats):
                quotient_dict[cand] = Fraction(n_votes, divisor)
        candidates, quotients = [
            list(x) for x in zip(*votelib.util.sorted_votes(
                quotient_dict, descending=False
            ))
        ]
        rem_seats = n_seats - sum(totals.values())
        while rem_seats > 0 and quotients:
            n_elect = 1
            max_q = quotients[-1]
            while n_elect < len(quotients) and quotients[-n_elect-1] == max_q:
                n_elect += 1
            to_elect = candidates[-n_elect:]
            if n_elect > rem_seats:
                to_elect = [votelib.evaluate.core.Tie(to_elect)] * rem_seats
            for cand in to_elect:
                if cand in totals:
                    totals[cand] += 1
                else:
                    totals[cand] = 1
                rem_seats -= 1
            for i in range(n_elect):
                cand = candidates.pop()
                quotients.pop()
                cand_total = totals.get(cand, 0)
                if cand_total < max_seats.get(cand, n_seats) and cand in votes:
                    new_quot = Fraction(
                        votes[cand], self.divisor_function(cand_total)
                    )
                    new_pos = bisect.bisect_left(quotients, new_quot)
                    candidates.insert(new_pos, cand)
                    quotients.insert(new_pos, new_quot)
        return {
            cand: cand_seats - prev_gains.get(cand, 0)
            for cand, cand_seats in totals.items()
            if cand_seats > prev_gains.get(cand, 0)
        }


@simple_serialization
class BiproportionalEvaluator:
    '''Allocate seats biproportionally to parties and constituencies.

    Biproportional apportionment is a method to provide proportional election
    results in two dimensions - constituencies and parties (candidates).
    It works by computing the proportional result using a highest averages
    method along one dimension (here: parties) and then iteratively updating it
    until proportionality is also reached for constituencies, in a process that
    somehow resembles iterative proportional fitting (IPF) but for integer
    values.

    There are two main biproportional apportionment algorithms - alternate
    scaling (AS), which is known to be faster in initial stages but also to
    stall in some corner cases, and tie-and-transfer (TT), which is slower but
    robust. Here, tie-and-transfer (TT) is implemented as described in [#puk]_.

    So far, the only studied variant of biproportional apportionment uses
    highest averages methods, but usually in an alternative specification by
    rounding rules (also called signpost sequences). These rounding rules have
    only been found for D'Hondt and Sainte-Laguë divisors; for other divisors,
    the implementation is missing yet.

    :param divisor_function: A callable producing the divisor from the number
        of seats awarded to the contestant so far, in the same form accepted
        by :class:`HighestAverages`.
    :param apportioner: An optional distribution evaluator to allocate seats to
        constituencies according to the total numbers of votes cast in each.
        Can also be an integer stating the uniformly valid number of seats for
        each constituency, or a dictionary giving the numbers per constituency.
        If None, the number of seats must be specified to :meth:`evaluate`,
        or the highest averages evaluator defined by *divisor_function* is
        used.
    :param signpost_q: The signpost function subtraction constant, which
        defines the rounding. For D'Hondt and Sainte-Laguë divisors, this is
        automatically determined; for other divisors, it must be specified
        manually.

    .. [#puk] "Chapter 15. Double-Proportional Divisor Methods:
        Technicalities", F. Pukelsheim. In: *Proportional Representation*,
        DOI ``10.1007/978-3-319-64707-4_15``.
    '''

    SIGNPOST_QS: Dict[str, Union[int, Fraction]] = {
        'd_hondt': 0,
        'sainte_lague': Fraction(1, 2),
    }

    def __init__(self,
                 divisor_function: Union[
                     str, Callable[[int], Number]
                 ] = 'd_hondt',
                 apportioner: Union[
                     votelib.evaluate.core.Distributor,
                     Dict[Constituency, int],
                     int,
                     None
                 ] = None,
                 signpost_q: Optional[Union[int, Fraction]] = None,
                 ):
        self.divisor_function = votelib.component.divisor.construct(
            divisor_function
        )
        if signpost_q is None:
            signpost_q = self._extract_signpost_q(self.divisor_function)
        self.signpost_q = signpost_q
        self.apportioner = apportioner
        self._eval = HighestAverages(self.divisor_function)

    def _extract_signpost_q(self,
                            fx: Callable[[int], Number],
                            ) -> Union[int, Fraction]:
        '''Determine the signpost subtraction constant from the divisor.'''
        value = self.SIGNPOST_QS.get(fx.__name__, NotImplemented)
        if value is NotImplemented:
            raise NotImplementedError(
                f'signpost q not known for {fx.__name__}, give it explicitly'
            )
        return value

    def evaluate(self,
                 votes: Dict[Constituency, Dict[Candidate, int]],
                 n_seats: Union[int, Dict[Constituency, int]],
                 ) -> Dict[Constituency, Dict[Candidate, int]]:
        '''Distribute seats biproportionally.

        :param votes: Simple votes per constituency to be evaluated.
        :param n_seats: Number of seats to be filled, either in total or by
            constituency.
        '''
        # Initial result, proportional by parties only.
        # All subsequent modifications preserve this proportionality.
        result = self._initial_solution(votes, n_seats)
        tgt_district_seats = votelib.evaluate.core.apportion(
            votes, n_seats,
            self.apportioner if self.apportioner is not None else self._eval,
        )
        # Initial coefficients (inverse divisors) for parties and districts.
        # These are modified by the tie-and-transfer algorithm.
        district_coefs = {d: 1 for d in votes}
        # Party coefficients are computed to be consistent with the initial
        # party-proportional seat allocation result.
        party_coefs = self._initial_party_coefs(votes, result)
        # Iterate the tie-and-transfer algorithm.
        while True:
            cur_district_seats = votelib.convert.ConstituencyTotals().convert(
                result
            )
            # Get districts that have less or more seats than needed.
            districts_under, districts_over = self._districts_unsat(
                cur_district_seats,
                tgt_district_seats,
            )
            if not (districts_under or districts_over):
                # Biproportionality achieved, terminate.
                return result
            quotients = self._calc_quots(votes, district_coefs, party_coefs)
            # Attempt to find a seat transfer path from a district with higher
            # than proportional seat count to a district with lower than
            # proportional count along cells with tied results while keeping
            # party totals.
            districts_labeled, parties_labeled = self._labeled(
                quotients, result, districts_under, districts_over
            )
            # If any undervalued district was reached by the path,
            districts_under_labeled = list(sorted(
                d for d in districts_under if d in districts_labeled
            ))
            if districts_under_labeled:
                # transfer the seat along that path.
                self._augment_result(
                    result, districts_labeled, parties_labeled,
                    districts_under_labeled[0], districts_over
                )
            else:
                # Otherwise, adjust some district and party coefficients
                # so that more ties are created along which the seats can be
                # transferred.
                adj_coef = self._adj_coef(
                    quotients,
                    result,
                    districts_labeled.keys(),
                    parties_labeled.keys()
                )
                if adj_coef == 0 or adj_coef >= 1:
                    raise votelib.evaluate.core.VotingSystemError(
                        f'invalid adjustment coefficient {adj_coef}'
                    )
                for district in districts_labeled:
                    district_coefs[district] *= adj_coef
                for party in parties_labeled:
                    party_coefs[party] /= adj_coef

    @staticmethod
    def _augment_result(result: Dict[Constituency, Dict[Candidate, int]],
                        districts_labeled: Dict[Constituency, Set[Candidate]],
                        parties_labeled: Dict[Candidate, Set[Constituency]],
                        start_district: Constituency,
                        districts_over: List[Constituency],
                        ) -> None:
        '''Transfer a seat to increase proportionality in result.

        Move one allocated seat to *start_district* along a path determined
        by alternating values in *districts_labeled* and *parties_labeled*
        through alternated additions and subtractions until a seat is
        subtracted from one of *districts_over*, which lowers the flaw count
        (disproportionality) by two seats.
        '''
        aug_path = [start_district]
        cur_source = districts_labeled
        while aug_path[-1] not in districts_over:
            aug_path.append(cur_source[aug_path[-1]].pop())
            if cur_source is parties_labeled:
                cur_source = districts_labeled
            else:
                cur_source = parties_labeled
        for i, ctup in enumerate(zip(aug_path[:-1], aug_path[1:])):
            if i % 2:
                party, district = ctup
                result[district][party] -= 1
                if not result[district][party]:
                    del result[district][party]
            else:
                district, party = ctup
                if party not in result[district]:
                    result[district][party] = 0
                result[district][party] += 1

    def _adj_coef(self,
                  quotients: Dict[Constituency, Dict[Candidate, Fraction]],
                  result: Dict[Constituency, Dict[Candidate, int]],
                  districts_labeled: Collection[Constituency],
                  parties_labeled: Collection[Candidate],
                  ) -> Fraction:
        '''Determine the adjustment coefficient that will create more ties.

        Seats can only be transferred along cells (district-party combinations)
        with a tied result. We aim to find a coefficient to multiply the cell
        quotient values in some columns or rows so that more ties are created.
        '''
        alpha = 0
        beta = INF
        for district, d_quots in quotients.items():
            d_is_labeled = district in districts_labeled
            if parties_labeled or d_is_labeled:
                for party, pd_quot in d_quots.items():
                    if d_is_labeled != (party in parties_labeled):
                        pd_seats = result[district].get(party, 0)
                        pd_signpost = pd_seats - self.signpost_q
                        is_alpha_scalable = (
                            d_is_labeled
                            and party not in parties_labeled
                            and pd_signpost > 0
                        )
                        if is_alpha_scalable:
                            pd_alpha = pd_signpost / pd_quot
                            if pd_alpha > alpha:
                                alpha = pd_alpha
                        is_beta_scalable = (
                            not d_is_labeled
                            and party in parties_labeled
                            and pd_quot > 0
                        )
                        if is_beta_scalable:
                            pd_beta = (pd_signpost + 1) / pd_quot
                            if pd_beta < beta:
                                beta = pd_beta
        return alpha if (alpha >= 1 / beta) else (1 / beta)

    def _labeled(self,
                 quotients: Dict[Constituency, Dict[Candidate, Fraction]],
                 result: Dict[Constituency, Dict[Candidate, int]],
                 districts_under: List[Constituency],
                 districts_over: List[Constituency],
                 ) -> Tuple[
                     Dict[Constituency, Set[Candidate]],
                     Dict[Candidate, Set[Constituency]]
                 ]:
        '''Attempt to find a seat transfer path along tied cells.'''
        all_parties = list(sorted(frozenset(
            p for dqs in quotients.values() for p in dqs.keys()
        )))
        # Start with all districts with higher values than needed.
        labeled_districts = collections.defaultdict(
            set, {d: set() for d in districts_over}
        )
        labeled_parties = collections.defaultdict(set)
        prev_n_labelings = -1
        n_labelings = 0
        # Repeat expanding the paths until a step produces no more labels.
        while prev_n_labelings < n_labelings:
            prev_n_labelings = n_labelings
            for d in labeled_districts:
                for party in all_parties:
                    if party not in labeled_parties:
                        is_downgradable = self._is_downgradable(
                            quotients[d][party],
                            result[d].get(party, 0)
                        )
                        if is_downgradable:
                            labeled_parties[party].add(d)
                            n_labelings += 1
            for party in labeled_parties:
                for d in quotients.keys():
                    if d not in labeled_districts:
                        is_upgradable = self._is_upgradable(
                            quotients[d][party],
                            result[d].get(party, 0)
                        )
                        if is_upgradable:
                            labeled_districts[d].add(party)
                            n_labelings += 1
            # If any district with lower value than needed was reached, the
            # path is complete.
            if any(d in districts_under for d in labeled_districts):
                break
        return labeled_districts, labeled_parties

    def _is_upgradable(self, quotient: Fraction, n_seats: int) -> bool:
        '''Check if the cell contains a tie and a seat can be added.'''
        return (
            int(quotient) == quotient - self.signpost_q
            and n_seats + 1 - self.signpost_q == quotient
        )

    def _is_downgradable(self, quotient: Fraction, n_seats: int) -> bool:
        '''Check if the cell contains a tie and a seat can be subtracted.'''
        return (
            int(quotient) == quotient - self.signpost_q
            and n_seats - self.signpost_q == quotient
            and n_seats >= 1
        )

    @staticmethod
    def _calc_quots(votes: Dict[Constituency, Dict[Candidate, int]],
                    district_coefs: Dict[Constituency, int],
                    party_coefs: Dict[Candidate, Fraction],
                    ) -> Dict[Constituency, Dict[Candidate, Fraction]]:
        '''Calculate fractional seat count approximators in both dimensions.

        Calculates from vote counts and coefficients (inverse divisors).
        '''
        return {
            district: {
                party: n_votes * district_coefs[district] * party_coefs[party]
                for party, n_votes in district_votes.items()
            }
            for district, district_votes in votes.items()
        }

    @staticmethod
    def _districts_unsat(cur_district_seats: Dict[Constituency, int],
                         tgt_district_seats: Dict[Constituency, int],
                         ) -> Tuple[List[Constituency], List[Constituency]]:
        '''Return districts with less and more seats than needed, respectively.

        If both are empty, proportionality is achieved.
        '''
        all_districts = (
            frozenset(cur_district_seats)
            | frozenset(tgt_district_seats)
        )
        under, over = [], []
        for d in all_districts:
            cur_d_seats = cur_district_seats.get(d, 0)
            tgt_d_seats = tgt_district_seats.get(d, 0)
            if cur_d_seats != tgt_d_seats:
                (under if cur_d_seats < tgt_d_seats else over).append(d)
        return under, over

    def _initial_solution(self,
                          votes: Dict[Constituency, Dict[Candidate, int]],
                          n_seats: Union[int, Dict[Constituency, int]],
                          ) -> Dict[Constituency, Dict[Candidate, int]]:
        '''Allocate seats proportionally along the party dimension.'''
        # First, allocate the total seats to parties.
        party_seats = self._eval.evaluate(
            votelib.convert.VoteTotals().convert(votes),
            n_seats if isinstance(n_seats, int) else sum(n_seats.values())
        )
        # Compute initial assignment through evaluation by party (columnwise)
        # to districts.
        solution = {d: {} for d in votes.keys()}
        for party, n_party_seats in party_seats.items():
            party_result = self._eval.evaluate(
                {district: votes[district].get(party, 0)
                 for district in votes},
                n_party_seats
            )
            for district, n_district_party_seats in party_result.items():
                if isinstance(district, votelib.evaluate.core.Tie):
                    # Tie on evaluation start, select an arbitrary district
                    # of the tied.
                    sel_district = list(sorted(district))[0]
                    solution[sel_district].setdefault(party, 0)
                    solution[sel_district][party] += n_district_party_seats
                else:
                    solution[district][party] = n_district_party_seats
        return solution

    def _initial_party_coefs(self,
                             votes: Dict[Constituency, Dict[Candidate, int]],
                             seats: Dict[Constituency, Dict[Candidate, int]],
                             ) -> Dict[Candidate, Fraction]:
        '''Determine initial party coefficients from their votes and seats.

        This is done to transpose the result obtained by conventional
        uniproportional evaluation to the biproportional format that used
        dimensional coefficients (inverse divisors). The initial form thus must
        be consistent with the seat counts awarded to parties by the initial
        evaluation. The coefficients will usually not differ much, only by the
        degree by which the initial solution is disproportional to parties.
        '''
        party_coefs = {}
        for party in votelib.convert.VoteTotals().convert(votes).keys():
            lowcoef = 0
            highcoef = INF
            for district, district_votes in votes.items():
                party_district_n_votes = district_votes.get(party, 0)
                if party_district_n_votes:
                    party_district_n_seats = seats[district].get(party, 0)
                    party_district_lowcoef = Fraction(
                        party_district_n_seats - self.signpost_q,
                        party_district_n_votes
                    )
                    if party_district_lowcoef > lowcoef:
                        lowcoef = party_district_lowcoef
                    party_district_highcoef = Fraction(
                        party_district_n_seats + 1 - self.signpost_q,
                        party_district_n_votes
                    )
                    if party_district_highcoef < highcoef:
                        highcoef = party_district_highcoef
            party_coefs[party] = Fraction(lowcoef + highcoef, 2)
        return party_coefs



"""Measure proportionality of election results.

These functions evaluate how proportionally the seats are allocated to parties
according to their votes received. For a review of such indicators, see
[#polrep]_ or [#kalog]_.

The functions only accept votes in simple format. To evaluate
disproportionality for elections using other vote types, you must use an
appropriate converter first; however, using the index for those election
systems might not be meaningful.

The seat counts must be in a dictionary format as returned by votelib's
distributors.

.. [#polrep] "polrep: Calculate Political Representation Scores", Didier
    Ruedin. https://rdrr.io/rforge/polrep/
.. [#kalog] "Measures of disproportionality", Kalogirou.
    http://www2.stat-athens.aueb.gr/~jpan/diatrives/Kalogirou/chapter5.pdf
"""


import math
from typing import Dict, Tuple
from numbers import Number

from votelib.candidate import Candidate


def gallagher(votes: Dict[Candidate, Number],
              results: Dict[Candidate, Number],
              ) -> float:
    """Compute the Gallagher index of election result disproportionality.

    The Gallagher (LSq) index [#lsq]_ expresses the mismatch between the
    fraction of votes received and seats allocated for each candidate or party.
    The index ranges from zero (no disproportionality) to 1 (total
    disproportionality).
    Compared to the Loosemore–Hanby index, it highlights large deviations
    rather than small ones.

    :param votes: Numbers of votes for each candidate.
    :param results: Seat counts awarded to each candidate.

    .. [#lsq] "Gallagher index", Wikipedia.
        https://en.wikipedia.org/wiki/Gallagher_index
    """
    paired_fractions = _vote_seat_fractions(votes, results)
    return math.sqrt(.5 * sum(
        (vote_frac - seat_frac) ** 2
        for vote_frac, seat_frac in paired_fractions.values()
    ))


def loosemore_hanby(votes: Dict[Candidate, Number],
                    results: Dict[Candidate, Number],
                    ) -> float:
    """Compute the Loosemore–Hanby index of election result disproportionality.

    The Loosemore–Hanby (LH) index [#lhind]_ expresses the mismatch between the
    fraction of votes received and seats allocated for each candidate or party.
    The index ranges from zero (no disproportionality) to 1 (total
    disproportionality).
    Compared to the Gallagher index, it does not diminish the effect of smaller
    deviations.

    :param votes: Numbers of votes for each candidate.
    :param results: Seat counts awarded to each candidate.

    .. [#lhind] "Loosemore–Hanby index", Wikipedia.
        https://en.wikipedia.org/wiki/Loosemore%E2%80%93Hanby_index
    """
    paired_fractions = _vote_seat_fractions(votes, results)
    return .5 * sum(
        abs(vote_frac - seat_frac)
        for vote_frac, seat_frac in paired_fractions.values()
    )


def rose(votes: Dict[Candidate, Number],
         results: Dict[Candidate, Number],
         ) -> float:
    """Compute the Rose disproportionality index (inverse LH). [#kalog]_

    This is an inverted version of the Loosemore-Hanby index which ranges from
    1 (no disproportionality) to 0 (total disproportionality).

    :param votes: Numbers of votes for each candidate.
    :param results: Seat counts awarded to each candidate.
    """
    return 1 - loosemore_hanby(votes, results)


def rae(votes: Dict[Candidate, Number],
        results: Dict[Candidate, Number],
        ) -> float:
    """Compute Rae's index of disproportionality. [#kalog]_

    Rae's index is the earliest known disproportionality measure. It is known
    to underestimate disproportionality in the presence of small parties.

    :param votes: Numbers of votes for each candidate.
    :param results: Seat counts awarded to each candidate.
    """
    paired_fractions = _vote_seat_fractions(votes, results)
    return sum(
        abs(vote_frac - seat_frac)
        for vote_frac, seat_frac in paired_fractions.values()
    ) / len(paired_fractions)


def lijphart(votes: Dict[Candidate, Number],
             results: Dict[Candidate, Number],
             ) -> float:
    """Compute the Lijphart's index of disproportionality. [#kalog]_

    Lijphart's index takes the single largest difference between vote and seat
    fractions.

    :param votes: Numbers of votes for each candidate.
    :param results: Seat counts awarded to each candidate.
    """
    paired_fractions = _vote_seat_fractions(votes, results)
    return max(
        abs(vote_frac - seat_frac)
        for vote_frac, seat_frac in paired_fractions.values()
    )


def sainte_lague(votes: Dict[Candidate, Number],
                 results: Dict[Candidate, Number],
                 ) -> float:
    """Compute the Sainte-Laguë index of disproportionality. [#kalog]_

    Sainte-Laguë index takes fractional differences.

    :param votes: Numbers of votes for each candidate.
    :param results: Seat counts awarded to each candidate.
    """
    paired_fractions = _vote_seat_fractions(votes, results)
    return sum(
        (vote_frac - seat_frac) ** 2 / vote_frac
        for vote_frac, seat_frac in paired_fractions.values()
    )


def d_hondt(votes: Dict[Candidate, Number],
            results: Dict[Candidate, Number],
            ) -> float:
    """Compute the D'Hondt index of disproportionality. [#kalog]_

    This is the index of disproportionality minimized by the D'Hondt highest
    averages evaluator. It uses the maximum ratio of seats to votes across
    parties.

    :param votes: Numbers of votes for each candidate.
    :param results: Seat counts awarded to each candidate.
    """
    paired_fractions = _vote_seat_fractions(votes, results)
    return max(
        seat_frac / vote_frac
        for vote_frac, seat_frac in paired_fractions.values()
    )


def regression(votes: Dict[Candidate, Number],
               results: Dict[Candidate, Number],
               ) -> float:
    """Compute the regression index of disproportionality. [#kalog]_

    This is obtained by performing linear regression to predict seat fractions
    from vote fractions. If the index is one, the allocation is perfectly
    proportional; values below one signal preference of the system for smaller
    parties, while values above one signal preference for larger parties.

    :param votes: Numbers of votes for each candidate.
    :param results: Seat counts awarded to each candidate.
    """
    num = 0
    denom = 0
    for vote_frac, seat_frac in _vote_seat_fractions(votes, results).values():
        num += vote_frac * seat_frac
        denom += vote_frac ** 2
    return num / denom


def _vote_seat_fractions(votes: Dict[Candidate, Number],
                         results: Dict[Candidate, Number],
                         ) -> Dict[Candidate, Tuple[Number, Number]]:
    total_votes = sum(votes.values())
    total_seats = sum(results.values())
    merged = {
        cand: (n_votes / total_votes, results.get(cand, 0) / total_seats)
        for cand, n_votes in votes.items()
    }
    for cand, result in results.items():
        if cand not in merged:
            merged[cand] = (0, result)
    return merged



'''General voting system evaluator machinery.'''

from __future__ import annotations

import abc
import collections
import inspect
from fractions import Fraction
from typing import Any, List, Dict, Union, Optional, Callable
from numbers import Number

import votelib.convert
import votelib.util
import votelib.vote
import votelib.component.quota
from votelib.candidate import Candidate, Constituency, Person, ElectionParty
from votelib.persist import simple_serialization


class VotingSystemError(Exception):
    '''A voting system with a valid setup ended up in an unresolvable state.'''
    pass


class Tie(frozenset):
    '''Candidates tied for a seat.

    This object, a subclass of ``frozenset``, is produced by evaluators that do
    not resolve all evaluation ties - for example, a plurality evaluator when
    two or more candidates have an equal number of votes and only some of them
    fit into the number of seats to fill. It can either be presented in the
    result (such as in the case when the given voting system has no tiebreaking
    mechanism) or caught by a tiebreaking evaluator - use :class:`TieBreaking`
    for this.

    This object also provides some special static and class methods to handle
    ties in some basic ways.
    '''
    @classmethod
    def reconcile(cls,
                  elected: List[Union[Candidate, Tie]]
                  ) -> List[Candidate]:
        '''Reconcile multiply tied candidates.

        A placeholder implementation to account for the situation where e.g.
        partial evaluations gave ties but their multiple occurrence in the
        overall result means the ties can be unambiguously resolved.

        :param elected: Selection result - a list of candidates or ties.
        '''
        n_places = collections.defaultdict(float)
        for i, result in enumerate(elected):
            if isinstance(result, cls):
                fraction = Fraction(1, len(result))
                for cand in result:
                    n_places[cand] += fraction
        if any(val >= 1 for val in n_places.values()):
            raise NotImplementedError('tie reconciliation')
            # select tied candidates with n_places over 1, put them in as
            # elected and subset the remaining tie objects
        else:
            # no change
            return elected

    @staticmethod
    def any(result: List[Union[Candidate, Tie]]) -> bool:
        '''Return True if there is any tie in the list, False otherwise.'''
        return any(isinstance(item, Tie) for item in result)

    @staticmethod
    def tie_rankings(rankings: List[List[Candidate]]
                     ) -> List[Union[Candidate, Tie]]:
        '''Form a single ranking out of a list of rankings.

        Produces ties where the rankings do not agree.
        '''
        raise NotImplementedError

    @classmethod
    def break_by_list(cls,
                      elected: List[Union[Candidate, Tie]],
                      breaker: List[Candidate],
                      ) -> List[Candidate]:
        '''Break ties in the elected list according to ordering in breaker.'''
        broken = []
        ties = {}
        for item in elected:
            if isinstance(item, Tie):
                if item in ties:
                    broken.append(ties[item][0])
                    if len(ties[item]) > 1:
                        ties[item] = ties[item][1:]
                    else:
                        del ties[item]
                else:
                    sorted_item = list(sorted(item, key=breaker.index))
                    broken.append(sorted_item[0])
                    ties[item] = sorted_item[1:]
            else:
                broken.append(item)
        return broken


def get_n_best(votes: Dict[Candidate, Number],
               n_seats: int,
               ) -> List[Union[Candidate, Tie]]:
    '''Return n_seats candidates with the highest number of votes.

    Essentially a plurality selection function.
    Produces ties correctly so is useful as a component in many other systems
    that use selection by maximum somewhere in their process.

    :param votes: Mapping of candidates to the number of votes obtained.
    :param n_seats: Number of seats to be filled.
    :returns: A list of top n_seats candidates. If there is a tie, the last
        items will refer to a single Tie object containing the tied candidates.
    '''
    sorted_items = votelib.util.sorted_votes(votes)
    if len(sorted_items) > n_seats:
        # find if there is a tie between the last elected and first unelected
        threshold_votes = sorted_items[n_seats-1][1]
        if sorted_items[n_seats][1] == threshold_votes:
            # tie detected, find all tied
            tied = []
            n_untied = None
            for i, item in enumerate(sorted_items):
                cand, n_votes = item
                if n_votes == threshold_votes:
                    tied.append(cand)
                    if n_untied is None:
                        n_untied = i
            n_tie_places = n_seats - n_untied
            return (
                [item[0] for item in sorted_items[:n_untied]]
                + [Tie(tied)] * n_tie_places
            )
        else:
            return [cand for cand, n_votes in sorted_items[:n_seats]]
    else:
        return [cand for cand, n_votes in sorted_items]


class Evaluator(metaclass=abc.ABCMeta):
    '''Evaluate votes for candidates and allocate seats to them.

    A root abstract base class for all evaluators.
    '''
    @abc.abstractmethod
    def evaluate(self, votes, *args, **kwargs
                 ) -> Union[List[Candidate], Dict[Candidate, int]]:
        '''Evaluate votes for candidates and allocate seats to them.'''
        raise NotImplementedError


class UnknownEvaluator(Evaluator):
    def evaluate(self, votes, *args, **kwargs
                 ) -> Union[List[Candidate], Dict[Candidate, int]]:
        raise NotImplementedError


class Selector(Evaluator):
    '''Elect a given number of candidates.

    Requires a number of seats to determine the number of candidates to elect.
    '''
    @abc.abstractmethod
    def evaluate(self, votes, n_seats, *args, **kwargs) -> List[Candidate]:
        '''Elect n_seats candidates as a list.

        :param votes: Votes of any type.
        :param n_seats: Number of candidates to elect.
        :returns: A list of candidates elected, ordered by magnitude of victory
            (winner first).
        '''
        raise NotImplementedError


class SeatlessSelector(Evaluator):
    '''Elect some candidates.

    Does not allow to pass a number of seats to determine the number of
    candidates to elect; it arises naturally from the votes and other selector
    settings.
    '''
    @abc.abstractmethod
    def evaluate(self, votes, *args, **kwargs) -> List[Candidate]:
        '''Elect a list of candidates without a guaranteed count.

        :param votes: Votes of any type.
        :returns: A list of candidates elected, ordered by magnitude of victory
            (winner first).
        '''
        raise NotImplementedError


class Distributor(Evaluator):
    '''Allocate seats to candidates based on collective preference.'''

    @abc.abstractmethod
    def evaluate(self,
                 votes: Dict[Any, int],
                 n_seats: int,
                 prev_gains: Dict[Candidate, int] = {},
                 max_seats: Dict[Candidate, int] = {},
                 ) -> Dict[Candidate, int]:
        '''Allocate n_seats to candidates as a dictionary.

        :param votes: Votes of any type.
        :param n_seats: Number of seats to allocate to candidates.
        :param prev_gains: Seats gained by the candidate/party in previous
            election rounds.
        :param max_seats: Maximum number of seats that the given
            candidate/party can obtain in total (including previous gains).
        :returns: Numbers of seats allocated to respective candidates.
            Candidates with no allocated seats do not appear in the dictionary.
            The ordering of the dictionary is unspecified.
        '''
        raise NotImplementedError


class SeatlessDistributor(Evaluator):
    '''Give seats to candidates based on collective preference.

    The number of seats cannot be specified - it stems from the nature of the
    election system.
    '''

    @abc.abstractmethod
    def evaluate(self,
                 votes: Dict[Any, int],
                 prev_gains: Dict[Candidate, int] = {},
                 max_seats: Dict[Candidate, int] = {},
                 ) -> Dict[Candidate, int]:
        '''Determine numbers of seats given to candidates, as a dictionary.

        :param votes: Votes of any type.
        :param prev_gains: Seats gained by the candidate/party in previous
            election rounds.
        :param max_seats: Maximum number of seats that the given
            candidate/party can obtain in total (including previous gains).
        :returns: Numbers of seats allocated to respective candidates.
            Candidates with no allocated seats do not appear in the dictionary.
            The ordering of the dictionary is unspecified.
        '''
        raise NotImplementedError


class SeatCountCalculator:
    def calculate(self,
                  votes: Dict[Any, int],
                  n_seats: int,
                  prev_gains: Dict[Candidate, int],
                  max_seats: Dict[Candidate, int] = {},
                  ) -> int:
        raise NotImplementedError


class OpenListEvaluator(metaclass=abc.ABCMeta):
    '''An abstract class for open list selection evaluators.

    Apart from votes and a number of seats, also requires the specification
    of a list of candidates to serve as the default (party-determined closed
    list) ordering to be overridden by preferential votes for candidates.
    '''
    @abc.abstractmethod
    def evaluate(self,
                 votes: Dict[Any, Number],
                 n_seats: int,
                 candidate_list: List[Candidate],
                 ) -> List[Candidate]:
        '''Elect n_seats candidates as a list.

        :param votes: Votes of any type.
        :param n_seats: Number of candidates to elect.
        :param candidate_list: Candidates in the order determined by their
            party.
        :returns: A list of candidates elected, ordered by magnitude of victory
            (winner first).
        '''
        raise NotImplementedError


@simple_serialization
class MultistageDistributor:
    '''A distribution evaluator with several rounds of awarding seats.

    Useful for several systems, e.g. multi-member proportional systems where
    the first round evaluates the constituencies and the second one evaluates
    the national level.

    :param rounds: Partial distributors. Will be called one by one, with the
        subsequent distributors getting the results of the previous steps as
        previous gains.
    :param depth: Nesting depth of the results. Use numbers larger than 1 when
        the results from the rounds are nested by constituency levels (2 for
        one constituency level, etc.)
    '''

    def __init__(self, rounds: List[Distributor], depth: int = 1):
        self.rounds = rounds
        self.depth = depth

    def evaluate(self,
                 votes: Union[Dict[Any, int], List[Dict[Any, int]]],
                 n_seats: int,
                 prev_gains: Dict[Candidate, int] = {},
                 max_seats: Dict[Candidate, int] = {},
                 ) -> Dict[Candidate, int]:
        '''Evaluate all rounds of the distribution.

        :param votes: Votes to be evaluated. Either a single set to be passed
            to all rounds, or separate sets of votes for each round.
        :param n_seats: Number of seats to be filled in total.
        :param prev_gains: Seats gained by the candidate/party in previous
            election rounds.
        :param max_seats: Maximum number of seats that the given
            candidate/party can obtain in total (including previous gains).
        '''
        elected = prev_gains.copy()
        if hasattr(votes, 'items'):
            votes = [votes] * len(self.rounds)
        for stage, stage_votes in zip(self.rounds, votes):
            stage_res = stage.evaluate(
                stage_votes, n_seats, prev_gains=elected, max_seats=max_seats
            )
            self._add_stage_results(elected, stage_res, self.depth)
        return elected

    def _add_stage_results(self, elected, stage_res, depth):
        if depth == 1:
            votelib.util.add_dict_to_dict(elected, stage_res)
        else:
            for constituency in set(elected.keys()) | set(stage_res.keys()):
                self._add_stage_results(
                    elected.setdefault(constituency, {}),
                    stage_res.get(constituency, {}),
                    depth - 1
                )


@simple_serialization
class UnusedVotesDistributor(MultistageDistributor):
    """Run several rounds of awarding seats, subtracting used votes.

    In some elections, candidates who receive a seat have a certain quota
    of votes (usually the quota used to grant that seat) subtracted from their
    vote count, which then continues into subsequent rounds of evaluation.
    An example is the new (after 2021) Czech Parliament lower house election.

    :param rounds: Partial distributors. Will be called one by one, with the
        subsequent distributors getting the results of the previous steps as
        previous gains and the votes progressively lowered according to
        corresponding subtracted quotas.
    :param quota_functions: Quota functions to use for used vote subtraction.
        The length of this list should be one item smaller than the number
        of rounds (since after the last round, there is no meaningful
        subtraction).
    :param depth: Nesting depth of the results. Use numbers larger than 1 when
        the results from the rounds are nested by constituency levels (2 for
        one constituency level, etc.)
    """
    def __init__(self,
                 rounds: List[Distributor],
                 quota_functions: Optional[List[
                     Union[str, Callable[[int, int], Number]]
                 ]] = None,
                 depth: int = 1,
                 ):
        super().__init__(rounds, depth)
        if quota_functions is None:
            try:
                quota_functions = [r.quota_function for r in self.rounds[:-1]]
            except AttributeError:
                raise ValueError('quota_functions not given but election'
                                 'rounds do not have own quota functions')
        self.quota_functions = [
            votelib.component.quota.construct(quota_def)
            for quota_def in quota_functions
        ]

    def evaluate(self,
                 votes: Dict[Candidate, int],
                 n_seats: int,
                 prev_gains: Dict[Candidate, int] = {},
                 max_seats: Dict[Candidate, int] = {},
                 ) -> Dict[Candidate, int]:
        elected = prev_gains.copy()
        if max_seats:
            raise NotImplementedError('max_seats not supported')
        for stage, quota_fx in zip(self.rounds, self.quota_functions + [None]):
            stage_res = stage.evaluate(votes, n_seats)
            self._add_stage_results(elected, stage_res, self.depth)
            if quota_fx is not None:    # if we have more steps to go
                votes = self._use_votes(
                    votes, stage_res, n_seats, quota_fx, depth=self.depth
                )
                n_seats = self._subtract_gained_seats(
                    n_seats, stage_res, depth=self.depth
                )
        return elected

    @classmethod
    def _use_votes(cls,
                   votes: Dict[Candidate, int],
                   elected: Dict[Candidate, int],
                   n_seats: int,
                   quota_fx: Callable[[int, int], Number],
                   depth: int = 1,
                   ) -> Dict[Candidate, int]:
        if depth == 1:
            quota_val = quota_fx(sum(votes.values()), n_seats)
            rem_votes = {}
            for cand, n_votes in votes.items():
                to_subtract = quota_val * elected.get(cand, 0)
                if n_votes < to_subtract:
                    raise VotingSystemError(f'would subtract {to_subtract}'
                                            f'votes but only has {n_votes}'
                                            f'for {cand}')
                rem_votes[cand] = n_votes - to_subtract
            return rem_votes
        else:
            if not isinstance(n_seats, dict):
                n_seats = collections.defaultdict(lambda: n_seats)
            return {
                constituency: cls._use_votes(
                    con_votes,
                    elected.get(constituency, {}),
                    n_seats.get(constituency, 0),
                    quota_fx=quota_fx,
                    depth=depth - 1
                )
                for constituency, con_votes in votes.items()
            }

    @classmethod
    def _subtract_gained_seats(cls,
                               n_seats: Union[int, Dict[Candidate, int]],
                               elected: Dict[Candidate, int],
                               depth: int = 1,
                               ) -> Union[int, Dict[Candidate, int]]:
        if isinstance(n_seats, dict):
            return {
                constituency: cls._subtract_gained_seats(
                    con_n_seats, elected.get(constituency, {}), depth=depth - 1
                )
                for constituency, con_n_seats in n_seats.items()
            }
        else:
            return n_seats - cls._gained_seats(elected, depth=depth)

    @classmethod
    def _gained_seats(cls,
                      elected: Dict[Any, Union[Number, Dict[Any, Any]]],
                      depth: int = 1,
                      ) -> int:
        if depth == 1:
            return sum(elected.values())
        else:
            return sum(
                cls._gained_seats(val, depth=depth - 1)
                for val in elected.values()
            )


@simple_serialization
class AdjustedSeatCount:
    '''Distribute an adjusted total number of seats according to votes cast.

    Useful in multi-member proportional systems where overhang seats are
    accounted for, e.g. by leveling.

    :param calculator: A calculator determining how many seats to award. It is
        called with the evaluation parameters (votes, intended number of seats,
        previously gained seats...).
    :param evaluator: A distribution evaluator producing the actual results
        with the adjusted number of seats.
    '''
    def __init__(self,
                 calculator: SeatCountCalculator,
                 evaluator: Distributor,
                 ):
        self.calculator = calculator
        self.evaluator = evaluator

    def evaluate(self,
                 votes: Dict[Any, int],
                 n_seats: int,
                 prev_gains: Dict[Candidate, int],
                 max_seats: Dict[Candidate, int] = {},
                 ) -> Dict[Candidate, int]:
        '''Distribute an adjusted total number of seats.

        :param votes: Votes to be evaluated, of any type accepted by the
            calculator and evaluator.
        :param n_seats: Intended (baseline) number of seats that would arise
            if no adjustment was necessary.
        :param prev_gains: Seats gained by the candidate/party in previous
            election rounds.
        :param max_seats: Maximum number of seats that the given
            candidate/party can obtain in total (including previous gains).
        '''
        seat_adj = self.calculator.calculate(
            votes, n_seats, prev_gains=prev_gains, max_seats=max_seats
        )
        return self.evaluator.evaluate(
            votes,
            n_seats + seat_adj,
            prev_gains=prev_gains,
            max_seats=max_seats,
        )


@simple_serialization
class AllowOverhang:
    '''Increase the total number of seats to allow keeping overhang seats.

    Overhang seats arise in multi-member proportional systems when a party
    gains more seats in the first round (usually local and plurality-based)
    than its proportional share according to second round (national,
    proportional party-based) votes. This calculator increases the number
    of seats so that all parties keep their overhang seats while the number of
    seats awarded in the second round stays constant. This leads to parties
    that performed strong in the first round but weak in the second round
    gaining more representation than would be decided by the second round only.

    This is the system used in New Zealand legislative elections.

    :param evaluator: A distribution evaluator producing the proportional
        results from the second round votes to determine which seats are
        overhang. Usually will be the same or very similar to the evaluator
        used in the second round directly.
    '''
    def __init__(self, evaluator: Distributor):
        self.evaluator = evaluator

    def calculate(self,
                  votes: Dict[Any, int],
                  n_seats: int,
                  prev_gains: Dict[Candidate, int],
                  max_seats: Dict[Candidate, int] = {},
                  ) -> int:
        '''Return an adjustment to the total number of seats to allow overhang.

        :param votes: Second round votes to produce the proportional result to
            determine overhang.
        :param n_seats: Intended (baseline) number of seats that would arise
            if no adjustment was necessary. This is passed to the internal
            evaluator to determine the proportional result.
        :param prev_gains: Seats gained in the first round results.
        :param max_seats: Maximum number of seats that the given
            candidate/party can obtain in total, including first round results.
        :returns: Zero if no adjustment is necessary, a positive integer if it
            is, amounting to the number of overhang seats detected.
        '''
        prop_result = self.evaluator.evaluate(
            votes, n_seats, max_seats=max_seats,
        )
        adj = 0
        for cand, n_prev_gained in prev_gains.items():
            prop_cand = prop_result.get(cand, 0)
            if prop_cand < n_prev_gained:
                adj += n_prev_gained - prop_cand
        return adj


@simple_serialization
class LevelOverhang:
    '''Increase the total number of seats to proportional, keeping overhang.

    Overhang seats arise in multi-member proportional systems when a party
    gains more seats in the first round (usually local and plurality-based)
    than its proportional share according to second round (national,
    proportional party-based) votes. This calculator increases the number
    of seats so that all parties keep their overhang seats while the overall
    seat counts for parties remain proportional according to the second round
    votes. This means that even when the first round seats end up dominated by
    a single party which has a huge overhang, the other parties still get
    enough *leveling* seats on the national level to maintain a representation
    proportional to their second round votes.

    :param evaluator: A distribution evaluator producing the proportional
        results from the second round votes to determine which seats are
        overhang. Usually will be the same or very similar to the evaluator
        used in the second round directly.
    '''
    def __init__(self, evaluator: Distributor):
        self.evaluator = evaluator

    def calculate(self,
                  votes: Dict[Any, int],
                  n_seats: int,
                  prev_gains: Dict[Candidate, int],
                  max_seats: Dict[Candidate, int] = {},
                  ) -> int:
        '''Return an adjustment to the total number of seats to level overhang.

        :param votes: Second round votes to produce the proportional result to
            determine overhang.
        :param n_seats: Intended (baseline) number of seats that would arise
            if no adjustment was necessary. This is passed to the internal
            evaluator to determine the proportional result.
        :param prev_gains: Seats gained in the first round results.
        :param max_seats: Maximum number of seats that the given
            candidate/party can obtain in total, including first round results.
        :returns: Zero if no adjustment is necessary, a positive integer if it
            is, amounting to the sum of overhang seats detected and leveling
            seats required.
        '''
        prop_result = self.evaluator.evaluate(
            votes, n_seats, max_seats=max_seats,
        )
        lowest_allowed = {
            party: max(prev_gains.get(party, 0), prop_gain)
            for party, prop_gain in prop_result.items()
        }
        nonprop_drop = 0
        for party, prev_gain in prev_gains.items():
            if party not in lowest_allowed:
                nonprop_drop += prev_gain
        adj_count = n_seats - nonprop_drop
        pmins = list(lowest_allowed.items())
        while any(prop_result[party] < minimum for party, minimum in pmins):
            adj_count += 1
            prop_result = self.evaluator.evaluate(
                votes, adj_count, max_seats=max_seats,
            )
        return adj_count + nonprop_drop - n_seats


@simple_serialization
class LevelOverhangByConstituency:
    '''Increase the total number of seats to proportional in constituencies.

    This is a variant :class:`LevelOverhang` that works out overhang and
    leveling seats on a constituency level (such as Länder in Germany)
    while still maintaining nationwide proportionality. For a detailed
    description of the overhang leveling process, see that class. This
    variant detects overhang seats for each constituency separately and
    progressively increases the number of leveling seats until proportionality
    is satisfied on a national level.

    This is the system used for the election to German Bundestag.

    :param constituency_evaluator: Evaluates the total party result per
        constituency.
    :param overall_evaluator: Evaluates the total party result nationwide;
        if not given, an aggregate of the constituency result is used
        instead.
    '''
    def __init__(self,
                 constituency_evaluator: Distributor,
                 overall_evaluator: Optional[Distributor] = None,
                 ):
        self.constituency_evaluator = constituency_evaluator
        self.overall_evaluator = overall_evaluator

    def calculate(self,
                  votes: Dict[Constituency, Dict[Any, int]],
                  n_seats: int,
                  prev_gains: Dict[Constituency, Dict[Candidate, int]],
                  max_seats: Dict[Constituency, Dict[Candidate, int]] = {},
                  ) -> int:
        '''Adjust the seat count to level overhang by constituency.

        :param votes: Second round votes to produce the proportional result to
            determine overhang, by constituency.
        :param n_seats: Intended (baseline) number of seats that would arise
            if no adjustment was necessary. This is passed to the internal
            evaluator to determine the proportional result.
        :param prev_gains: Seats gained in the first round results by parties,
            by constituency.
        :param max_seats: Maximum number of seats that the given
            candidate/party can obtain in total, including first round results,
            by constituency.
        :returns: Zero if no adjustment is necessary, a positive integer if it
            is, amounting to the sum of overhang seats detected and leveling
            seats required for all constituencies.
        '''
        # Get lowest amount of seats per party and constituency as the maximum
        # of the party's first round count and its theoretical proportional
        # second round result.
        cty_results = self.constituency_evaluator.evaluate(
            votes, n_seats, max_seats=max_seats,
        )
        lowest_allowed = votelib.convert.VoteTotals().convert({
            cty: {
                party: max(prev_gains.get(cty, {}).get(party, 0), prop_seats)
                for party, prop_seats in cty_prop_seats.items()
            }
            for cty, cty_prop_seats in cty_results.items()
        })
        # If any party did not make it to the second round, subtract
        # its first round result from the total number of seats when
        # determining the proportional national result, add them back at the
        # end.
        nonprop_drop = 0
        for cty_gains in prev_gains.values():
            for party, prev_gain in cty_gains.items():
                if party not in lowest_allowed:
                    nonprop_drop += prev_gain
        adj_count = n_seats - nonprop_drop
        overall_evaluator = self.overall_evaluator
        if overall_evaluator:
            party_votes = votelib.convert.VoteTotals().convert(votes)
        else:
            overall_evaluator = votelib.convert.PostConverted(
                self.constituency_evaluator,
                votelib.convert.DistributionMerger()
            )
            party_votes = votes
        # Progressively increase the number of seats until all parties get at
        # least to the minimum given for each of its constituencies.
        prop_result = overall_evaluator.evaluate(
            party_votes, adj_count, max_seats=max_seats
        )
        while any(prop_result.get(party, 0) < minimum
                  for party, minimum in lowest_allowed.items()):
            adj_count += 1
            # Slow since it re-runs the whole proportional evaluation, but
            # I haven't found a general analytical solution finding the overall
            # result directly.
            prop_result = overall_evaluator.evaluate(
                party_votes, adj_count, max_seats=max_seats,
            )
        return adj_count + nonprop_drop - n_seats


@simple_serialization
class PostConverted:
    '''An evaluator whose results are run through a converter.

    Useful when the evaluator is a part of a larger system and its output needs
    to be adapted to it, e.g. in multi-member proportional systems where first
    votes are received by candidates in a selection evaluation but the second
    votes are evaluated by distribution to parties.

    :param evaluator: An evaluator to run.
    :param converter: A converter to apply on the results of the evaluator.
    '''
    def __init__(self, evaluator, converter):
        self.evaluator = evaluator
        self.converter = converter

    def evaluate(self, votes, *args, **kwargs):
        '''Run the evaluator and return its result, converted.

        All other arguments are passed through to the evaluator.

        :param votes: Votes for the evaluator.
        :param n_seats: Number of seats to allocate by the evaluator.
        '''
        return self.converter.convert(
            self.evaluator.evaluate(votes, *args, **kwargs)
        )


@simple_serialization
class PreConverted:
    '''An evaluator whose votes are first run through a converter.

    Useful when the evaluator accepts a different type of votes than the
    actual ballots, where the converter adapts them - e.g. when the votes are
    by constituency but the evaluation is nationwide, or when ranked votes are
    given but a part of the system runs on simple votes.

    :param evaluator: An evaluator to run.
    :param converter: A converter to apply on the votes before passing them to
        the evaluator.
    '''
    def __init__(self, converter, evaluator):
        self.converter = converter
        self.evaluator = evaluator

    def evaluate(self, votes, *args, **kwargs):
        '''Convert the votes by and evaluate them through the evaluator.

        All other arguments are passed through to the evaluator.

        :param votes: Votes to be passed to the converter.
        :param n_seats: Number of seats to allocate by the evaluator.
        '''
        conv_votes = self.converter.convert(votes)
        return self.evaluator.evaluate(conv_votes, *args, **kwargs)


DEFAULT_SUBSETTER = votelib.vote.SimpleSubsetter()


@simple_serialization
class Conditioned:
    '''An evaluator whose votes are pre-selected to exclude some variants.

    Before passing the votes to the main evaluator, an eliminator is evaluated
    first, and only the candidates returned by it are allowed to proceed to the
    main evaluation. This is useful for implementing vote thresholds in
    proportional systems.

    :param evaluator: The main evaluator to produce the results.
    :param eliminator: A selector to determine which variants proceed to the
        main evaluator - only those that are returned by its evaluate method
        do so. It can accept previous gain counts but should not need
        the number of seats (it should be independent of it).
    :param subsetter: A subsetter to subset a vote to just concern the
        candidates returned by the eliminator.
    '''
    def __init__(self,
                 eliminator: SeatlessSelector,
                 evaluator: Evaluator,
                 subsetter: Optional[votelib.vote.SimpleSubsetter] = None,
                 depth: int = 1,
                 ):
        self.eliminator = eliminator
        self.evaluator = evaluator
        if subsetter is None:
            subsetter = DEFAULT_SUBSETTER
        self.subsetter = subsetter
        self.depth = depth

    def evaluate(self,
                 votes: Dict[Any, int],
                 n_seats: Optional[int] = None,
                 prev_gains: Dict[Candidate, int] = {},
                 **kwargs) -> Union[List[Candidate], Dict[Candidate, int]]:
        '''Evaluate the main evaluator for variants that passed the eliminator.

        All other keyword arguments are passed through to the main evaluator.

        :param votes: Votes to be passed to the main evaluator and eliminator.
        :param n_seats: Number of seats to allocate by the main evaluator.
        :param prev_gains: Numbers of seats previously gained by the candidates
            (e.g. in previous scrutinia). Will be passed to both the main
            evaluator and eliminator, if they accept them (as determined by
            the `accepts_seats` function).
        '''
        summed_votes = self._sum_party_votes(votes, depth=self.depth)
        summed_prev_gains = self._sum_party_votes(prev_gains, depth=self.depth)
        if accepts_prev_gains(self.eliminator):
            not_eliminated = self.eliminator.evaluate(
                summed_votes, prev_gains=summed_prev_gains
            )
        else:
            not_eliminated = self.eliminator.evaluate(summed_votes)
        elim_votes = self._elim_party_votes(
            votes, not_eliminated, depth=self.depth
        )
        if accepts_prev_gains(self.evaluator):
            kwargs = kwargs.copy()
            kwargs['prev_gains'] = prev_gains
        if accepts_seats(self.evaluator):
            return self.evaluator.evaluate(
                elim_votes, n_seats, **kwargs
            )
        else:
            return self.evaluator.evaluate(elim_votes, **kwargs)

    @classmethod
    def _sum_party_votes(cls,
                         values: Dict[Any, Dict[Any, Any]],
                         depth: int = 1,
                         ) -> Dict[Candidate, Any]:
        if depth == 1:
            return values
        else:
            return votelib.convert.VoteTotals().convert({
                key: cls._sum_party_votes(val, depth=depth - 1)
                for key, val in values.items()
            })

    def _elim_party_votes(self,
                          votes: Dict[Any, Dict[Any, Any]],
                          not_eliminated: List[Candidate],
                          depth: int = 1,
                          ) -> Dict[Candidate, Any]:
        if depth == 1:
            return votelib.convert.SubsettedVotes(self.subsetter).convert(
                votes, not_eliminated
            )
        else:
            return {
                key: self._elim_party_votes(
                    val, not_eliminated, depth=depth - 1
                )
                for key, val in votes.items()
            }


@simple_serialization
class ByConstituency:
    '''Perform constituency-wise evaluation of given system.

    Evaluates and returns the results of the given system separately per each
    constituency it is given votes for. If desired, apportions the seats to
    constituencies according to the total numbers of votes cast in each.

    The selector and vote subsetter perform what the :class:`Conditioned`
    evaluator wrapper would do; however, they are only used after apportionment
    is performed, because that is often defined before candidates are excluded.
    If this is not desired, define the selector and vote subsetter on a
    :class:`Conditioned` wrapper, with the selector wrapped in an additional
    :class:`PreConverted` with :class:`votelib.convert.VoteTotals`,
    and leave the selector undefined here.

    :param evaluator: Evaluator to use on the constituency level.
    :param apportioner: An optional distribution evaluator to allocate seats to
        constituencies according to the total numbers of votes cast in each.
        Can also be an integer stating the uniformly valid number of seats for
        each constituency, or a dictionary giving the numbers per constituency.
        If None, the number of seats must be specified to :meth:`evaluate`.
    :param preselector: An optional selection evaluator to select candidates
        eligible for constituency-wise evaluation. Votes for candidates that
        do not pass its selection will be removed for all constituencies before
        evaluation. If not given, no preselection will be applied.
    :param subsetter: A subsetter to subset a vote to just concern the
        candidates returned by the selector. The default option needs to be
        modified if the votes are more deeply nested.
    '''
    def __init__(self,
                 evaluator: Evaluator,
                 apportioner: Union[
                     Distributor, Dict[Constituency, int], int, None
                 ] = None,
                 preselector: Optional[Selector] = None,
                 subsetter: votelib.vote.VoteSubsetter = DEFAULT_SUBSETTER,
                 ):
        self.evaluator = evaluator
        self.apportioner = apportioner
        self.preselector = preselector
        self.subsetter = subsetter

    def evaluate(self,
                 votes: Dict[Constituency, Dict[Any, int]],
                 n_seats: Union[int, Dict[Constituency, int], None] = None,
                 prev_gains: Dict[Constituency, Dict[Candidate, int]] = {},
                 max_seats: Dict[Constituency, Dict[Candidate, int]] = {},
                 ) -> Union[
                     Dict[Constituency, Dict[Candidate, int]],
                     Dict[Constituency, List[Candidate]],
                 ]:
        '''Return the election results evaluated by constituency.

        :param votes: Votes in the format accepted by the inner evaluator.
        :param n_seats: Number of seats to be allocated:

            -   If ``None``, the apportioner must be a distributor that does
                not accept seat count, and will be called with the total
                numbers of votes cast in each constituency to provide a seat
                count for each constituency.
            -   If an integer and the apportioner is None, regarded as the
                uniform number of seats for each constituency.
            -   If an integer and the apportioner is a distributor, the
                apportioner will be called with this integer as the total
                number of seats to distribute it to the constituencies.
            -   If a dictionary, it will be regarded as the mapping of
                constituencies to their numbers of seats.

        :param prev_gains: Seats gained by the candidate/party in previous
            election rounds in each constituency, to inform the underlying
            evaluator.
        :param max_seats: Maximum number of seats that the given
            candidate/party can obtain in total per constituency
            (including previous gains).
        :returns: Results of the evaluation by constituency.
        '''
        apportionment = apportion(
            votes, n_seats, apportioner=self.apportioner
        )
        preselected = self._preselect(votes, n_seats)
        results = {}
        no_value_districts = []
        for district, dvotes in votes.items():
            result = self._evaluate_district(
                dvotes,
                apportionment.get(district),
                preselected,
                prev_gains.get(district, {}),
                max_seats.get(district, {}),
            )
            if result is None:
                no_value_districts.append(district)
            else:
                results[district] = result
        result_type = type(next(iter(results.values())))
        for district in no_value_districts:
            results[district] = result_type()
        return results

    def _evaluate_district(self,
                           votes: Dict[Any, Number],
                           n_seats: int,
                           preselected: List[Candidate],
                           prev_gains: Dict[Candidate, int],
                           max_seats: Dict[Candidate, int],
                           ) -> Union[Dict[Candidate, int], List[Candidate]]:
        if n_seats == 0:
            return None
        else:
            if preselected is not None:
                votes = votelib.convert.SubsettedVotes(self.subsetter).convert(
                    votes, preselected
                )
            if accepts_prev_gains(self.evaluator):
                return self.evaluator.evaluate(
                    votes, n_seats,
                    prev_gains=prev_gains,
                    max_seats=max_seats,
                )
            else:
                return self.evaluator.evaluate(votes, n_seats)

    def _preselect(self, votes, n_seats):
        if self.preselector:
            # select candidates that passed national level conditions
            nat_agg = votelib.convert.VoteTotals()
            nat_votes = nat_agg.convert(votes)
            if accepts_seats(self.preselector):
                return self.preselector.evaluate(nat_votes, n_seats)
            else:
                return self.preselector.evaluate(nat_votes)
        else:
            return None


@simple_serialization
class PreApportioned:
    """Apportion seats to constituencies before the evaluation.

    In some systems, apportionment is not fixed but is only done at the time
    of the election based on the number of votes cast in the constituencies.
    This component also allows to perform this.

    :param evaluator: The per-constituency election evaluator.
    :param apportioner: How to apportion the seats to constituencies:

        -   If an integer, each constituency gets this many seats.
        -   If a dictionary mapping constituencies to integers, each
            constituency gets the number of seats specified.
        -   If a distribution evaluator, it is called on vote totals in each
            constituency to produce the number of seats. In this variant,
            n_seats must be passed to ``evaluate()``.
    """
    def __init__(self,
                 evaluator: Distributor,
                 apportioner: Union[Distributor, Dict[Constituency, int], int],
                 ):
        self.evaluator = evaluator
        self.apportioner = apportioner

    def evaluate(self,
                 votes: Dict[Constituency, Dict[Any, int]],
                 n_seats: Union[int, Dict[Constituency, int], None] = None,
                 prev_gains: Dict[Constituency, Dict[Candidate, int]] = {},
                 max_seats: Dict[Constituency, Dict[Candidate, int]] = {},
                 ) -> Union[
                     Dict[Constituency, Dict[Candidate, int]],
                     Dict[Constituency, List[Candidate]],
                 ]:
        apportionment = apportion(
            votes, n_seats, apportioner=self.apportioner
        )
        return self.evaluator.evaluate(
            votes,
            n_seats=apportionment,
            prev_gains=prev_gains,
            max_seats=max_seats,
        )


@simple_serialization
class RemovedApportionment:
    """Disregard previously computed apportionment for this evaluator.

    In some systems, only a single (usually the "proportionality leveling")
    component does not respect an otherwise valid apportionment; this wrapper
    sums the apportioned seats back up into a single number for it.

    :param evaluator: The wrapped evaluator that expects unapportioned seats
        (i.e. a single number).
    """
    def __init__(self, evaluator: Distributor):
        self.evaluator = evaluator

    def evaluate(self,
                 votes: Dict[Constituency, Dict[Any, int]],
                 n_seats: Union[int, Dict[Constituency, int], None] = None,
                 prev_gains: Dict[Constituency, Dict[Candidate, int]] = {},
                 max_seats: Dict[Constituency, Dict[Candidate, int]] = {},
                 ) -> Union[
                     Dict[Constituency, Dict[Candidate, int]],
                     Dict[Constituency, List[Candidate]],
                 ]:
        return self.evaluator.evaluate(
            votes,
            n_seats=sum(n_seats.values()),
            prev_gains=prev_gains,
            max_seats=max_seats,
        )


@simple_serialization
class ByParty:
    '''Distribute overall party results among its constituency lists.

    This is an inverted variant of the :class:`ByConstituency` evaluator.
    Here, the total seat counts for each party are determined by an overall
    evaluator on nationally aggregated votes and an allocation distributor is
    subsequently used to disaggregate the result of the party to
    constituencies.

    Useful for some mixed-member proportional system such as the German
    Bundestag one, where this mechanism is used to distribute overhang leveling
    seats.

    :param overall_evaluator: Evaluator to use on the central (national) level.
    :param allocator: An optional distribution evaluator to allocate seats of
        the party to individual constituencies. If None, the overall evaluator
        is reused. In either case, the allocator must accept simple votes
        for constituencies as candidates.
    :param subsetter: A subsetter according to the vote type used, to
        extract votes for any single party from the overall votes.
    '''
    def __init__(self,
                 overall_evaluator: Distributor,
                 allocator: Optional[Distributor] = None,
                 subsetter: votelib.vote.VoteSubsetter = DEFAULT_SUBSETTER,
                 ):
        self.overall_evaluator = overall_evaluator
        self.allocator = allocator
        self.subsetter = subsetter

    def evaluate(self,
                 votes: Dict[Constituency, Dict[Any, int]],
                 n_seats: Optional[int] = None,
                 prev_gains: Dict[Constituency, Dict[Candidate, int]] = {},
                 max_seats: Dict[Constituency, Dict[Candidate, int]] = {},
                 ) -> Dict[Constituency, Dict[Candidate, int]]:
        '''Return constituency-wise election results evaluated by party.

        :param votes: Votes in the format accepted by the overall evaluator.
            They will be subsetted and aggregated to simple votes for
            constituencies for the allocator evaluation.
        :param n_seats: Number of seats to be allocated. If ``None``, the
            overall evaluator must be seatless.
        :param prev_gains: Seats gained by the candidate/party in previous
            election rounds in each constituency, to inform the allocator.
            Not passed to the overall evaluator.
        :param max_seats: Maximum number of seats that the given
            candidate/party can obtain in total per constituency
            (including previous gains). Not passed to the overall evaluator.
        :returns: Results of the evaluation by constituency.
        '''
        overall_votes = votelib.convert.VoteTotals().convert(votes)
        overall_result = self.overall_evaluator.evaluate(
            overall_votes, n_seats
        )
        allocator = self.allocator
        if allocator is None:
            allocator = self.overall_evaluator
        subset_conv = votelib.convert.SubsettedVotes(self.subsetter)
        results = collections.defaultdict(dict)
        for party, n_party_seats in overall_result.items():
            party_votes = {
                constituency: sum(
                    subset_conv.convert(cvotes, [party]).values()
                ) for constituency, cvotes in votes.items()
            }
            if accepts_prev_gains(allocator):
                party_prev_gains = {
                    constituency: cg[party]
                    for constituency, cg in prev_gains.items() if party in cg
                }
                party_max_seats = {
                    constituency: cm[party]
                    for constituency, cm in max_seats.items() if party in cm
                }
                allocated = allocator.evaluate(
                    party_votes, n_party_seats,
                    prev_gains=party_prev_gains, max_seats=party_max_seats
                )
            else:
                allocated = allocator.evaluate(party_votes, n_party_seats)
            for constituency, cseats in allocated.items():
                results[constituency][party] = cseats
        for constituency in votes.keys():
            if constituency not in results:
                results[constituency] = {}
        return dict(results)


@simple_serialization
class FixedSeatCount:
    '''An evaluator wrapper that provides a fixed seat count.

    Useful when the seat count for a given system is predefined and constant.
    Then, you do not need to specify it each time you call the evaluator, just
    provide the votes.

    :param evaluator: The evaluator to wrap. Must accept number of seats into
        its evaluate method.
    :param n_seats: The fixed number of seats to provide to the evaluator at
        each call.
    '''

    accepts_seats = False

    def __init__(self, evaluator: Evaluator, n_seats: int):
        if not accepts_seats(evaluator):
            raise ValueError(f'cannot wrap seatless evaluator {evaluator}'
                             f'into FixedSeatCount')
        self.evaluator = evaluator
        self.n_seats = n_seats

    def evaluate(self,
                 votes: Dict[Any, int],
                 **kwargs) -> Union[List[Candidate], Dict[Candidate, int]]:
        '''Run the inner evaluator with the predefined number of seats.

        All other keyword parameters are passed through to the wrapped
        evaluator.

        :param votes: Votes of the type accepted by the wrapped evaluator.
        '''
        return self.evaluator.evaluate(
            votes, self.n_seats, **kwargs
        )


@simple_serialization
class PartyListEvaluator:
    '''Evaluate which candidates are elected from party lists.

    Useful to determine the actual representatives elected when the result is
    first evaluated by party. This wrapper evaluates the party-based result
    first, then uses a party list evaluator to determine which candidates from
    the given party lists actually get elected.

    :param party_eval: A distribution evaluator determining the party-based
        election result.
    :param list_eval: An open list evaluator determining which candidates from
        the party list to elect, according to the preferential votes for the
        candidates. Look to the :mod:`votelib.evaluate.openlist` for some
        of these. If None, the party lists are considered closed and the
        elected candidates are taken from the top of each list, without
        considering the list votes.
    :param list_votes_converter: A converter to apply to list votes before
        passing to the list evaluator. The converter should produce a result
        where the candidate votes are grouped by party
        (such as :class:`votelib.convert.GroupVotesByParty`). If not given,
        the list votes are passed unchanged.
    '''
    def __init__(self,
                 party_eval: Distributor,
                 list_eval: Optional[OpenListEvaluator] = None,
                 list_votes_converter: Optional[
                     votelib.convert.Converter
                 ] = None,
                 ):
        self.party_eval = party_eval
        self.list_eval = list_eval
        self.list_votes_converter = list_votes_converter

    def evaluate(self,
                 votes: Dict[Any, Number],
                 n_seats: int, *,
                 party_lists: Dict[ElectionParty, List[Candidate]],
                 list_votes: Dict[ElectionParty, Dict[Any, Number]] = None,
                 **kwargs) -> Dict[ElectionParty, List[Person]]:
        '''Return lists of candidates elected for each party.

        All keyword arguments are passed to the party evaluator.

        :param votes: Votes in the format accepted by the party evaluator.
        :param n_seats: Number of seats to be allocated, to be passed to the
            party evaluator.
        :param party_lists: Lists of candidates for each party in the order
            they were submitted for election.
        :param list_votes: Votes for individual candidates on the party lists.
            If not given, the list evaluator must be None and the party lists
            are considered closed.
        :raises ValueError: If only one of list votes and a list evaluator
            is given.
        '''
        party_result = self.party_eval.evaluate(votes, n_seats, **kwargs)
        if self.list_eval is None:
            # evaluate closed lists
            if list_votes:
                raise ValueError('list votes given but no list evaluator')
            return {
                party: party_lists[party][:n_party_seats]
                for party, n_party_seats in party_result.items()
            }
        else:
            # evaluate open lists
            if not list_votes:
                raise ValueError('no list votes for open list evaluation')
            if self.list_votes_converter:
                list_votes = self.list_votes_converter.convert(list_votes)
            return {
                party: self.list_eval.evaluate(
                    list_votes[party], n_party_seats, party_lists[party]
                )
                for party, n_party_seats in party_result.items()
            }


def accepts_seats(evaluator: Evaluator) -> bool:
    '''Whether evaluator takes seat count as an argument to evaluate().'''
    if hasattr(evaluator, 'accepts_seats'):
        return evaluator.accepts_seats
    else:
        params = inspect.signature(evaluator.evaluate).parameters
        return 'n_seats' in params or _has_generic(params)


def accepts_prev_gains(evaluator: Evaluator) -> bool:
    '''Whether evaluator takes previous gains as an argument to evaluate().'''
    return 'prev_gains' in inspect.signature(evaluator.evaluate).parameters


def _has_generic(params: Dict[str, inspect.Parameter]) -> bool:
    return any(
        param.kind in (
            inspect.Parameter.VAR_POSITIONAL,
            inspect.Parameter.VAR_KEYWORD
        )
        for param in params.values()
    )


@simple_serialization
class Plurality:
    '''Plurality voting / maximum score evaluator. Elects a list of candidates.

    This encompasses the following voting systems:

    -   *First-past-the-post* (simple plurality, FPTP) for a single seat and
        one vote per person.
    -   *Single non-transferable vote* (SNTV) for multiple seats and one vote
        per voter.
    -   *Plurality-at-large* (multiple non-transferable vote) for
        multiple seats and number of votes equal the number of seats.
    -   *Limited voting* for multiple seats
        and number of votes per voter fixed and less than the number of seats.
    -   *Cumulative voting*, a very similar variant where the number of
        available votes per voter is not tied to the number of candidates
        or seats.
    -   *Approval voting* (AV) for number of available votes per voter equal
        to the number of candidates (but not the more advanced variants
        of approval voting such as proportional approval voting - see the
        :mod:`approval` for that). A multiwinner variant of approval voting
        is sometimes called block approval voting and exhibits very different
        properties. Use :class:`votelib.convert.ApprovalToSimpleVotes` to
        convert approval votes to simple votes accepted by this evaluator.
    -   *Satisfaction approval voting* (SAV), with one arbitrarily splittable
        vote per voter. Use :class:`votelib.convert.ApprovalToSimpleVotes`
        to convert approval votes to simple votes accepted by this evaluator.
    -   *Score voting* after the votes are aggregated by a score vote
        aggregator such as :class:`votelib.convert.ScoreToSimpleVotes`
        (:class:`votelib.evaluate.cardinal.ScoreVoting` can be used instead).

    These voting systems all use this evaluator together with some vote
    conversion or validation criteria.
    '''

    def evaluate(self,
                 votes: Dict[Candidate, Number],
                 n_seats: int = 1,
                 ) -> List[Candidate]:
        '''Select candidates by plurality voting.

        In case of a tie, returns :class:`Tie` objects at the end of the list
        of elected candidates.
        There will be as many ties as the number of tied seats;
        each tie will group the tied candidates.

        :param votes: Simple votes (mapping the candidates to a quantity, the
            more the better).
        :param n_seats: Number of candidates to select.
        :returns: A list of elected candidates, sorted in descending order by
            the input votes.
        '''
        return get_n_best(votes, n_seats)


@simple_serialization
class TieBreaking:
    '''Break ties from the main evaluation through a dedicated tiebreaker.

    Runs the main evaluator on the input, and if ties are present in its
    output, runs them through a separately specified selector with the
    original votes subsetted to just those concerning the tied candidates.

    Tiebreakers can be nested if there are more tiebreaking methods with
    different priority. To do so, wrap the core evaluator with the highest
    priority tiebreaker first, and then supply the result as the main evaluator
    to another tiebreaker with lower priority.

    :param main: The main evaluator; might be a selector or a distributor.
    :param tiebreaker: A selector to evaluate ties.
    :param vote_subsetter: A subsetter to subset a vote to just concern the
        tied candidates.
    '''
    def __init__(self,
                 main: Evaluator,
                 tiebreaker: Selector,
                 subsetter: votelib.vote.VoteSubsetter = DEFAULT_SUBSETTER,
                 ):
        self.main = main
        self.tiebreaker = tiebreaker
        self.subsetter = subsetter

    def evaluate(self, votes, *args, **kwargs):
        '''Evaluate the election, breaking ties if they arise.

        Any arguments besides the votes dictionary are passed unchanged to the
        main evaluator.

        :param votes: Votes for the election; will be used to feed the main
            evaluator (and the tiebreaker, after subsetting).
        '''
        main_result = self.main.evaluate(votes, *args, **kwargs)
        subset_conv = votelib.convert.SubsettedVotes(self.subsetter)
        if Tie.any(main_result):
            main_result = main_result.copy()
            ties = self._collect_ties(main_result)
            for tie, n_tied_seats in ties.items():
                sub_votes = subset_conv.convert(votes, tie)
                broken = self.tiebreaker.evaluate(sub_votes, n_tied_seats)
                if hasattr(main_result, 'items'):
                    self._replace_distr_ties(main_result, tie, broken)
                else:
                    self._replace_sel_ties(main_result, tie, broken)
        return main_result

    @staticmethod
    def _collect_ties(result: Union[List[Candidate], Dict[Candidate, int]]
                      ) -> Dict[Tie, int]:
        ties = collections.defaultdict(int)
        if hasattr(result, 'items'):
            for elected, n_seats in result.items():
                if isinstance(elected, Tie):
                    ties[elected] += n_seats
        else:
            for elected in result:
                if isinstance(elected, Tie):
                    ties[elected] += 1
        return ties

    @staticmethod
    def _replace_distr_ties(result: Dict[Candidate, int],
                            tie: Tie,
                            repl: List[Candidate],
                            ) -> None:
        del result[tie]
        for cand in repl:
            result[cand] = result.get(cand, 0) + 1

    @staticmethod
    def _replace_sel_ties(result: List[Candidate],
                          tie: Tie,
                          repl: List[Candidate],
                          ) -> None:
        for cand in repl:
            result[result.index(tie)] = cand


def apportion(votes: Dict[Constituency, Dict[Candidate, Number]],
              n_seats: Union[int, Dict[Constituency, int]],
              apportioner: Union[
                  Distributor, Dict[Constituency, int], int, None
              ] = None,
              ) -> Dict[Constituency, int]:
    if isinstance(apportioner, int):
        # fixed seats for each constituency
        return {c: apportioner for c in votes.keys()}
    elif hasattr(apportioner, 'items'):
        # seats per district are pre-determined statically
        return apportioner
    elif hasattr(n_seats, 'items'):
        # seats per district are determined dynamically outside this
        return n_seats
    elif apportioner:
        constituency_votes = {
            c: sum(c_votes.values()) for c, c_votes in votes.items()
        }
        if isinstance(n_seats, int):
            # apportion seats across districts by number of votes cast
            return apportioner.evaluate(constituency_votes, n_seats)
        elif n_seats is None:
            # apportionment without fixed total
            return apportioner.evaluate(constituency_votes)
        else:
            raise ValueError('unknown apportionment scheme')
    elif isinstance(n_seats, int):
        # delegate to subevaluator
        return {c: n_seats for c in votes.keys()}
    else:
        raise ValueError('invalid apportionment setup')



'''Evaluate the results of the elections.

There are two basic election types - selections and distributions.
In selections, each candidate (usually a physical person) is either
elected or not elected.
In distributions, some candidates (usually parties) are allocated
a positive number of seats (seats are distributied among the parties),
while other parties get none and do not appear in the result.

*Selection evaluators* return a list of candidates. If some of them are tied,
a :class:`core.Tie` object will appear in the list. The tie object will contain
all candidates tied for the seats and will be repeated the
number of times equal to the number of tied seats.

*Distribution evaluators* return a dictionary mapping candidates to the number
of seats (or other tokens they should be allocated). In case of a tie, the
:class:`core.Tie` object will appear as one of the keys, with the number of
tied seats as its associated value.

There are some other election concepts, for example *asset voting*, where the
candidates bargain with each other with the votes they have received;
these understandably cannot be implemented within the current scope of
Votelib. [#rvasset]_

None of the evaluators validate vote correctness; use the tools in the
:mod:`vote` module for that, potentially wrapped in
:class:`convert.InvalidVoteEliminator`.

.. [#rvasset] "Asset voting", Range voting.org.
    https://rangevoting.org/Asset.html
'''

from votelib.evaluate.core import *    # noqa



'''Advanced approval voting methods.

This module contains approval voting evaluators that cannot be reduced to
a plurality evaluation by aggregating the scores. Use
:class:`votelib.convert.ApprovalToSimpleVotes` in conjunction with
:class:`votelib.evaluate.Plurality` to evaluate simple approval voting (AV)
or satisfaction approval voting (SAV).
'''

import itertools
import collections
from fractions import Fraction
from numbers import Number
from typing import List, FrozenSet, Dict, Union, Callable

import votelib.evaluate.core
import votelib.component.quota
from votelib.candidate import Candidate
from votelib.persist import simple_serialization


@simple_serialization
class ProportionalApproval:
    '''Proportional Approval Voting (PAV) evaluator. [#wpav]_

    This method uses approval votes (voters select one or more permissible
    candidates) and evaluates the satisfaction of voters with each of the
    combinations of elected candidates. The satisfaction for each voter is
    given as the sum of reciprocals from 1 to N, where N is the number of
    elected candidates that the voter approved of.

    WARNING: Due to the enumeration of all candidate set combinations, this
    method is highly computationally expensive (``O(n!)`` in the number of
    candidates) and infeasible on common machines for more than a handful of
    candidates.

    Tie breaking not implemented - the method itself does not provide a way
    to do it, a dedicated tie breaker will probably be necessary.

    .. [#wpav] "Proportional approval voting", Wikipedia.
        https://en.wikipedia.org/wiki/Proportional_approval_voting
    '''
    def __init__(self):
        self._coefs = [0]

    def evaluate(self,
                 votes: Dict[FrozenSet[Candidate], int],
                 n_seats: int,
                 ) -> List[Candidate]:
        '''Select candidates by proportional approval.

        :param votes: Approval votes.
        :param n_seats: Number of candidates to be elected.
        :returns: Selected candidates in decreasing order measured by
            drop in satisfaction when the given candidate is excluded from the
            selected set.
        '''
        if len(self._coefs) < n_seats:
            self._coefs += [
                sum(Fraction(1, k + 1) for k in range(n))
                for n in range(len(self._coefs), n_seats + 1)
            ]
        best_alts = self._get_best_alternatives(votes, n_seats)
        if len(best_alts) == 1:
            return self._order_by_score(frozenset(best_alts[0]), votes)
        else:
            raise NotImplementedError(f'tied PAV alternatives: {best_alts}')
            # common = best_alts[0].intersection(*best_alts[1:])
            # return Tie.reconcile(self._order_by_score(common) +

    def _order_by_score(self,
                        alternative: FrozenSet[Candidate],
                        votes: Dict[FrozenSet[Candidate], int],
                        ) -> List[Candidate]:
        '''Order the candidates within an alternative.

        To output a correctly sorted list, we need to extend the PAV algorithm
        to impose an ordering to the set. This is done by sorting in decreasing
        order measured by drop in satisfaction when the given candidate is
        excluded from the selected set.
        '''
        satisfaction_drops = {
            cand: -self._satisfaction(alternative - {cand}, votes)
            for cand in alternative
        }
        return votelib.evaluate.core.get_n_best(
            satisfaction_drops, len(alternative)
        )

    def _get_best_alternatives(self,
                               votes: Dict[FrozenSet[Candidate], int],
                               n_seats: int,
                               ) -> List[FrozenSet[Candidate]]:
        '''Get the selection alternative(s) with the highest satisfaction.'''
        all_candidates = frozenset(
            cand for alt in votes.keys() for cand in alt
        )
        best_alternatives = []
        best_score = -float('inf')
        # evaluate each alternative
        for alternative in itertools.combinations(all_candidates, n_seats):
            # compute total satisfaction
            satisfaction = self._satisfaction(frozenset(alternative), votes)
            if satisfaction > best_score:
                best_alternatives = [alternative]
                best_score = satisfaction
            elif satisfaction == best_score:
                best_alternatives.append(alternative)
        return best_alternatives

    def _satisfaction(self,
                      alternative: FrozenSet[Candidate],
                      votes: Dict[FrozenSet[Candidate], int],
                      ) -> float:
        return sum(
            self._coefs[len(alt & alternative)] * n_votes
            for alt, n_votes in votes.items()
        )


@simple_serialization
class SequentialProportionalApproval:
    '''Sequential Proportional Approval Voting (SPAV) evaluator. [#wspav]_

    This method uses approval votes (voters select one or more permissible
    candidates) but evaluates them iteratively, unlike proportional approval
    voting. In each iteration, the best candidate is selected and all ballots
    that approve of them are reduced in value to ``1/n``, where ``n`` is the
    number of the candidates on that ballot already elected plus one (the
    value of those votes thus decreases to a half after one of the marked
    candidates is elected, to a third if a second one is elected, and so on).

    Tie breaking not yet implemented.

    .. [#wspav] "Sequential proportional approval voting", Wikipedia.
        https://en.wikipedia.org/wiki/Sequential_proportional_approval_voting
    '''
    def evaluate(self,
                 votes: Dict[FrozenSet[Candidate], int],
                 n_seats: int,
                 ) -> List[Candidate]:
        '''Select candidates by sequential proportional approval.

        :param votes: Approval votes.
        :param n_seats: Number of candidates to be elected.
        :returns: Selected candidates ordered as they were selected in the
            successive iterations.
        '''
        elected = []
        while len(elected) < n_seats:
            round_votes = collections.defaultdict(int)
            for cand_set, n_votes in votes.items():
                n_elected_from_set = len(cand_set.intersection(elected))
                for cand in cand_set:
                    round_votes[cand] += Fraction(
                        n_votes, n_elected_from_set + 1
                    )
            for cand in elected:
                del round_votes[cand]
            choice = votelib.evaluate.core.get_n_best(round_votes, 1)
            if not choice:
                return elected
            best = choice[0]
            if isinstance(best, votelib.evaluate.core.Tie):
                raise NotImplementedError('tie breaking in SPAV')
            else:
                elected.append(best)
        return elected


@simple_serialization
class QuotaSelector:
    '''Quota threshold (plurality) selector.

    Elects candidates with more (or also equally many, depending on
    *accept_equal*) votes than the specified quota.
    This often gives fewer candidates than the number of seats, and thus
    usually needs to be accompanied by an another evaluation step. In very rare
    cases, it might select more candidates than the number of seats.

    This is a component in the following systems:

    -   *Two-round runoff* (usually with the Droop quota and a single seat)
        where it gives the first-round winner if they have a majority of votes,
        and no one otherwise.

    It can also serve as a threshold evaluator (eliminator) in proportional
    systems that restrict the first party seat from being a remainder seat,
    or a kickstart for Huntington-Hill related methods that are not defined
    for zero-seat parties.

    :param quota_function: A callable producing the quota threshold from the
        total number of votes and number of seats.
    :param accept_equal: Whether to elect candidates that only just reach the
        quota threshold (this is known to produce some instabilities).
    :param on_more_over_quota: How to handle the case when more candidates
        fulfill the quota that there is seats:

        -   ``'error'``: raise a
            :class:`votelib.evaluate.core.VotingSystemError`,
        -   ``'select'``: select the candidates with the most votes (possibly
            producing ties when they are equal).
    '''
    def __init__(self,
                 quota_function: Union[
                     str, Callable[[int, int], Number]
                 ] = 'droop',
                 accept_equal: bool = True,
                 on_more_over_quota: str = 'error',
                 ):
        self.quota_function = votelib.component.quota.construct(quota_function)
        self.accept_equal = accept_equal
        self.on_more_over_quota = on_more_over_quota

    def evaluate(self,
                 votes: Dict[Candidate, int],
                 n_seats: int = 1,
                 ) -> List[Candidate]:
        qval = self.quota_function(
            sum(votes.values()), n_seats
        )
        over_quota = {}
        unselected = set()
        for cand, n_votes in votes.items():
            if n_votes > qval or self.accept_equal and n_votes == qval:
                over_quota[cand] = n_votes
            else:
                unselected.add(cand)
        if len(over_quota) > n_seats:
            if self.on_more_over_quota == 'error':
                raise votelib.evaluate.core.VotingSystemError(
                    f'wanted {n_seats}, quota gave {len(over_quota)}'
                )
            elif self.on_more_over_quota != 'select':
                raise ValueError(
                    f'invalid more_over_quota setting: {self.more_over_quota}'
                )
        return votelib.evaluate.core.get_n_best(over_quota, n_seats)



'''Divisor functions used in highest-averages proportional voting systems.

This provides arguments for the
:class:`votelib.evaluate.proportional.HighestAverages` evaluator. However,
divisors can be used as a component in many other voting systems,
such as biproportional allocation.

A divisor function takes the order number (usually equal to the number of
seats allocated so far) and returns the divisor by which to divide the number
of votes for the given party or candidate. The party or candidate with the
largest result then gets the next seat.

Some systems use a mathematically defined divisor but artificially change the
result for parties with no seats so far (`order == 0`) to make it harder for
parties to get their seats. Use :func:`modified_first_coef` for that.

All supported divisor functions are assembled in the `DIVISORS` dictionary
keyed by their name. `get()` retrieves from this dictionary by string key;
`construct()` also accepts callables and passes them through.
'''

from fractions import Fraction
from decimal import Decimal
from typing import Callable
from numbers import Number

import votelib.component.core


DIVISORS = {}


divisor_mark, get, construct = votelib.component.core.register_functions(
    DIVISORS, 'divisor', Callable[[int], Number]
)


@divisor_mark
def d_hondt(order: int) -> int:
    '''D'Hondt divisor, the most commonly used divisor.

    Forms a simple sequence 1, 2, 3...
    In the United States, this is known as the Jefferson divisor that was used
    for congressional apportionment 1792-1842.

    Known to slightly favor larger parties.
    '''
    return order + 1


@divisor_mark
def sainte_lague(order: int) -> int:
    '''Sainte-Laguë (Webster, Schepers) divisor, a commonly used divisor.

    Forms a sequence 1, 3, 5...

    Known to favor mid-sized parties.
    '''
    return 2 * order + 1


@divisor_mark
def imperiali(order: int) -> Fraction:
    '''Imperiali divisor. Not to be confused with the Imperiali quota.

    Forms a sequence 1, 1.5, 2...

    Known to favor large parties greatly.
    '''
    return Fraction(order, 2) + 1


@divisor_mark
def huntington_hill(order: int) -> Decimal:
    '''Huntington-Hill divisor.

    The divisors are defined by `sqrt(n*(n+1))`. This means the
    divisor is invalid for the zeroth order (passing a zero gives
    a divisor of zero, which raises an error in the subsequent division)
    and can thus only be used in cases where the first seat is already
    guaranteed (use the `prev_gains` argument for that).

    Used for United States congressional apportionment as of 2020.
    '''
    return Decimal(order * (order + 1)).sqrt()


@divisor_mark
def danish(order: int) -> int:
    '''Danish divisor.

    Forms a sequence 1, 4, 7...

    Extremely favors smaller parties.
    '''
    return 3 * order + 1


@divisor_mark
def macau(order: int) -> int:
    '''Macau modified D'Hondt divisor.

    This uses order counts as exponents (1, 2, 4, 8...), and therefore favors
    smaller parties.
    '''
    return 2 ** order


def modified_first_coef(divisor_fx: Callable[[int], Number],
                        first_coef: Decimal = Decimal('1.4'),
                        ) -> Callable[[int], Number]:
    '''Modify the divisor for the zeroth order to an apriori coefficient.

    This can be used to raise the threshold for parties that have not yet
    obtained a seat. This is the case in the Czech regional election
    (Koudelka coefficient 1.42), Nepal, Norway, and Sweden.

    :param divisor_fx: The ordinary divisor function to be wrapped and used for
        the first order and subsequent ones.
    :param first_coef: The coefficient to be used when order == 0.
    '''
    if not isinstance(first_coef, (int, Fraction)):
        first_coef = Fraction(*first_coef.as_integer_ratio())

    def _modified_divisor(order: int) -> Number:
        return divisor_fx(order) if order > 0 else first_coef

    return _modified_divisor



'''Open party list selection evaluators.

Many party-list proportional systems feature open lists - that is, voters get
to cast preferential votes for the candidates on that list, and the order of
those candidates might then be altered by these preferences. These evaluators
serve to select the candidates from the party list that actually get one of the
seats allocated to it.

The evaluators in this module accept an extra argument to their ``evaluate()``
methods, namely the original ordering of the candidates on the party list. Many
methods only alter this ordering (allow *jumps*) if the lower candidates have
significantly more votes than those above them; almost all of them rely on this
ordering to break ties.

For the most common open list evaluation case, :class:`ThresholdOpenList`
(allowing jumps when the candidate gets at least a given fraction of the total
votes for the list or at least a quota of the votes based on the number of
seats) is provided.
:class:`ListOrderTieBreaker` provides a wrapper for any
selection evaluator and uses the list ordering to break ties.
'''


from typing import Any, List, Dict, Union, Callable
from numbers import Number
from fractions import Fraction

import votelib.util
import votelib.component.quota
import votelib.evaluate.core
from votelib.candidate import Candidate
from votelib.persist import simple_serialization


@simple_serialization
class ThresholdOpenList:
    '''A threshold-based open list evaluator.

    Allows candidates that got at least a given number of votes for the list as
    preferential votes to jump over all other candidates. These jumping
    candidates are selected first, ordered by their number of votes; the rest
    of the seats is filled by other candidates on the list in its
    original order.

    The number of votes required for the jump is determined relative
    to the total number of votes received by the list (``jump_fraction``)
    or to a quota computed from this total and the number of seats to award
    to the list (``quota_function`` and ``quota_fraction``).

    :param jump_fraction: The fraction of total votes for the list that the
        candidate must obtain to jump ahead in the rankings.
    :param quota_function: A callable producing the quota threshold (number
        of votes required to jump) from the
        total number of votes and number of seats. The common quota functions
        can be referenced by string name from the
        :mod:`votelib.component.quota` module. The Hare quota (``'hare'``)
        is used most frequently.
    :param quota_fraction: A number to multiply the calculated quota with, to
        allow e.g. all candidates that got at least half of the quota to jump.
    :param take_higher: If both the jump fraction and quota are specified,
        this determines whether to take the higher of the thresholds (amounting
        to an AND function) or the lower one (amounting to an OR function, the
        default).
    :param accept_equal: Whether to use non-strict comparison for the number
        of votes against the quota or jump fraction.
    :param list_precedence: Whether the original list ordering takes precedence
        when more candidates are allowed to jump than the number of seats.
        If True, the candidates lowest on the list are eliminated; if False,
        the candidates with the least votes are eliminated.
    '''
    def __init__(self,
                 jump_fraction: Fraction = None,
                 quota_function: Union[
                     str, Callable[[int, int], Number], None
                 ] = None,
                 quota_fraction: Fraction = 1,
                 take_higher: bool = False,
                 accept_equal: bool = False,
                 list_precedence: bool = False,
                 ):
        self.jump_fraction = jump_fraction
        self.quota_fraction = quota_fraction
        if quota_function is not None and quota_fraction != 1:
            wrapped = votelib.component.quota.construct(quota_function)

            def _quota_fractional(votes: int, seats: int) -> Fraction:
                return wrapped(votes, seats) * quota_fraction

            self.quota_function = _quota_fractional
        else:
            self.quota_function = quota_function
        self.take_higher = take_higher
        self.accept_equal = accept_equal
        self.list_precedence = list_precedence

    def evaluate(self,
                 votes: Dict[Candidate, Number],
                 n_seats: int,
                 candidate_list: List[Candidate],
                 ) -> List[Candidate]:
        '''Select candidates from an open party list.

        :param votes: Preferential votes (simple) for candidates on the list.
        :param n_seats: Number of seats to be awarded to the list - the number
            of candidates to elect.
        :param candidate_list: The original (party-determined) list ordering.
        '''
        jump_thresholds = []
        total_votes = sum(votes.values())
        if self.jump_fraction is not None:
            jump_thresholds.append(total_votes * self.jump_fraction)
        if self.quota_function is not None:
            jump_thresholds.append(self.quota_function(total_votes, n_seats))
        if not jump_thresholds:
            return candidate_list[:n_seats]
        else:
            # selects all over (or equaling) the threshold, sorted by votes
            threshold = (max if self.take_higher else min)(jump_thresholds)
            jumping = [
                cand for cand, n_votes in votelib.util.sorted_votes(votes)
                if n_votes > threshold or (
                    self.accept_equal and n_votes == threshold
                )
            ]
            if len(jumping) > n_seats:
                if self.list_precedence:
                    # list ordering takes precedence, eliminate those lowest
                    jumping.sort(key=candidate_list.index)
                    jumping = jumping[:n_seats]
                    jumping.sort(key=votes.get, reverse=True)
                    return jumping
                else:
                    return jumping[:n_seats]
            else:
                # if seats left, take the highest on the list not yet elected
                elected = jumping
                for list_i, cand in enumerate(candidate_list):
                    if len(elected) == n_seats:
                        break
                    if cand not in elected:
                        elected.append(cand)
                return elected


@simple_serialization
class ListOrderTieBreaker:
    '''A wrapper for any selector for open-list candidate selection.

    The candidates that the provided evaluator returns with the given votes are
    provided as the result of the open list. If there is a tie, it is
    broken by the ordering on the party list.

    :param evaluator: Any selection evaluator.
    '''
    def __init__(self, evaluator: votelib.evaluate.core.Selector):
        self.evaluator = evaluator

    def evaluate(self,
                 votes: Dict[Any, Number],
                 n_seats: int,
                 candidate_list: List[Candidate],
                 ) -> List[Candidate]:
        '''Select candidates from an open party list.

        :param votes: Preferential votes for candidates on the list. Must be
            of the type that is accepted by the underlying evaluator.
        :param n_seats: Number of seats to be awarded to the list - the number
            of candidates to elect.
        :param candidate_list: The original (party-determined) list ordering.
        '''
        inner_result = self.evaluator.evaluate(votes, n_seats)
        if votelib.evaluate.core.Tie.any(inner_result):
            return votelib.evaluate.core.Tie.break_by_list(
                inner_result, candidate_list
            )
        else:
            return inner_result



'''Evaluators for special partial purposes, especially tiebreaking.

These evaluators should not be used as the main component of an election
system (except for obscure ones). Many of them choose the winners randomly,
so they are useful for tiebreaking, but that is about it.

You can make the random evaluators outputs stable if you give them a seed for
the random generator, but be careful with that in a real-world setting.
'''

import re
import random
import struct
import hashlib
import unicodedata
import operator
from typing import Any, List, Dict, Collection, Optional, Union
from numbers import Number
from decimal import Decimal

from votelib.candidate import Candidate
from votelib.persist import simple_serialization
import votelib.util


"""
class SpoiledBallotRemover:
    '''Returns only unspoiled ballots (votes for valid candidates).'''
    def evaluate(self,
                 votes: Dict[Candidate, Any],
                 ) -> List[Candidate]:
        return [
            cand for cand in votes.keys()
            if not isinstance(cand, NoneOfTheAbove)
                and not isinstance(cand, ReopenNominations)
        ]
"""


@simple_serialization
class RandomUnrankedBallotSelector:
    '''Select candidates by drawing random simple ballots from the tally.

    Useful for tiebreaking. Can also be used to evaluate *random approval
    voting* if the approval votes are converted to simple first.

    :param seed: Seed for the random generator that performs the sampling.
    '''
    def __init__(self,
                 seed: Optional[int] = None,
                 ):
        self.seed = seed
        self.stable = (self.seed is not None)

    def evaluate(self,
                 votes: Dict[Candidate, Number],
                 n_seats: int = 1,
                 ) -> List[Candidate]:
        '''Select candidates by drawing random ballots.

        :param votes: Simple votes.
        :param n_seats: Number of candidates (ballots) to be selected.
        '''
        random.seed(self.seed)
        return votelib.util.select_n_random(votes, n_seats)


@simple_serialization
class Sortitor:
    '''Perform sortition (random sampling) among the candidates.

    This selects the candidates purely randomly, with equal probabilities for
    each of them. Useful for tiebreaking and some obscure protocols such as
    Venetian doge election.

    :param seed: Seed for the random generator that performs the sampling.
    '''
    def __init__(self,
                 seed: Optional[int] = None,
                 ):
        self.seed = seed
        self.stable = (self.seed is not None)

    def evaluate(self,
                 votes: Dict[Candidate, Any],
                 n_seats: int = 1,
                 ) -> List[Candidate]:
        '''Select candidates randomly.

        :param votes: Simple votes. The quantities of votes are disregarded.
        :param n_seats: Number of candidates to be selected.
        '''
        random.seed(self.seed)
        return votelib.util.select_n_random({
            cand: 1 for cand, n_votes in votelib.util.sorted_votes(votes)
        }, n_seats)


@simple_serialization
class InputOrderSelector:
    '''Select first N candidates as they appear in the vote counts.

    This is useful for tiebreaking with an externally determined sort order,
    e.g. by ballot numbers or pre-generated random numbers. It takes advantage
    of dictionaries in Python 3.7+ maintaining insertion order.
    '''
    def evaluate(self,
                 votes: Dict[Candidate, Any],
                 n_seats: int = 1,
                 ) -> List[Candidate]:
        '''Select the first candidates that appear in the votes dictionary.

        :param votes: Simple votes. The quantities of votes are disregarded.
        :param n_seats: Number of candidates to be selected.
        '''
        return [cand for i, cand in enumerate(votes.keys()) if i < n_seats]


@simple_serialization
class CandidateNumberRanker:
    '''Select first N candidates with lowest candidate number.

    This is useful for tiebreaking with an externally determined sort order.
    '''
    def evaluate(self,
                 votes: Dict[Candidate, Any],
                 n_seats: int = 1,
                 ) -> List[Candidate]:
        '''Select the first candidates that appear in the votes dictionary.

        :param votes: Simple votes. The quantities of votes are disregarded.
        :param n_seats: Number of candidates to be selected.
        '''
        return list(
            sorted(votes.keys(), operator.attrgetter('number'))
        )[:n_seats]


@simple_serialization
class RFC3797Selector:
    '''Select candidates randomly by the algorithm from RFC 3797.

    This is a well-defined random selection method using external sources
    of randomness, that are to be provided as numbers or lists thereof.
    Once the sources of randomness are fixed in the constructor, the
    selection is deterministic with regard to the input order of candidates
    (candidate names or votes are disregarded).

    :param sources: Sources of randomness (seeds). For detailed
        recommendations on where to take them from, see the RFC. The
        list should contain an item per source. The list items may contain
        a number, a string (which will be stripped of accents and anything
        besides ASCII alphanumeric characters and uppercased) or a list of
        numbers. Strings and floats are STRONGLY not recommended.

    .. [#rfc3797] "RFC 3797: Publicly Verifiable Nominations Committee
        (NomCom) Random Selection", D. Eastlake 3rd.
        https://tools.ietf.org/html/rfc3797
    '''
    IMPURE_CHARS = re.compile('[^a-zA-Z0-9]')
    ORDER_STRUCT = struct.Struct('>H')
    UNPACK_STRUCT = struct.Struct('>Q')
    FIRST_PART_MULT = 1 << 64

    def __init__(self, sources: List[Union[Number, Collection[Number]]]):
        if not hasattr(hashlib, 'md5'):
            raise ImportError('need Python build with hashlib.md5()')
        self.sources = sources
        self.seed_bytes = self.source_bytestring(self.sources)

    def evaluate(self,
                 votes: Dict[Candidate, Any],
                 n_seats: int = 1,
                 ) -> List[Candidate]:
        '''Select the candidates by the algorithm from RFC 3797.

        :param votes: Simple votes. The quantities of votes are disregarded.
        :param n_seats: Number of candidates to be selected.
        '''
        cands = list(votes.keys())
        selected = []
        for i in range(n_seats):
            rand_int = self._random_value(i)
            selected.append(cands.pop(rand_int % len(cands)))
        return selected

    def _random_value(self, i: int) -> int:
        order_bytes = self.ORDER_STRUCT.pack(i)
        hash = hashlib.md5(
            order_bytes + self.seed_bytes + order_bytes
        ).digest()
        part1, part2 = self.UNPACK_STRUCT.iter_unpack(hash)
        return part1[0] * self.FIRST_PART_MULT + part2[0]

    @classmethod
    def source_bytestring(cls,
                          sources: List[Union[Number, Collection[Number]]]
                          ) -> bytes:
        '''Create the base randomness string from given randomness sources.

        Follows the procedure as given by Section 4 of the RFC.
        '''
        components = []
        for source in sources:
            if isinstance(source, str):
                component = cls._clean_string(source)
            elif hasattr(source, '__len__'):
                # collection of numbers, concatenate them ordered
                component = ''.join(
                    # order from smallest to largest
                    cls._num_to_str(num) for num in sorted(source)
                )
            else:
                # single item
                component = cls._num_to_str(source)
            # suffix by slash and concatenate as ascii
            components.append(component + '/')
        return ''.join(components).encode('ascii')

    @classmethod
    def _clean_string(cls, s: str) -> str:
        # rip off everything except ascii letters and numbers
        # and convert to uppercase
        deaccented = ''.join(
            c for c in unicodedata.normalize('NFD', s)
            if unicodedata.category(c) != 'Mn'
        )
        return cls.IMPURE_CHARS.sub('', deaccented).upper()

    @staticmethod
    def _num_to_str(num: Number) -> str:
        value = str(Decimal(num))
        if '.' not in value:
            value += '.'
        return value.rstrip('0')
